#include <QDir>
#include <QMapIterator>

#include "AutoImport.h"
#include "DialogImportNZB.h"
#include "mainwindow.h"

AutoImport::AutoImport(MainWindow *mainwindow, QNNTPGrabGlue *glue)
{
    this->glue = glue;
    this->mainwindow = mainwindow;

    connect(glue, SIGNAL(pluginEvent(QString,QString,QList<QString>)), SLOT(onPluginEvent(QString,QString,QList<QString>)));

    connect(&watcher, SIGNAL(directoryChanged(QString)), SLOT(directoryChanged(QString)));

    connect(&fileReadyTimer, SIGNAL(timeout()), SLOT(fileReadyTimer_timeout()));
}

void AutoImport::enableWatch(QString path)
{
    disableWatch();
    this->path = path;
    watcher.addPath(path);
}

void AutoImport::disableWatch()
{
    if (!this->path.isEmpty()) {
        watcher.removePath(this->path);
    }
}

void AutoImport::directoryChanged(QString path)
{
    /* Something changed in path, check if there are any .nzb files added to this folder */
    QDateTime stamp;
    QDir d;
    d.setPath(path);

    stamp = QDateTime::currentDateTime();
    stamp.addSecs(-10);

    QFileInfoList nzb_files = d.entryInfoList(QStringList("*.nzb"), QDir::Files | QDir::Readable);
    Q_FOREACH(QFileInfo nzb_file, nzb_files) {
        /* Only add files which were changed in the last 10 seconds */
        if (nzb_file.lastModified() < stamp) {
            continue;
        }

        /* Add the file to a queue. Every second a check will be performed to check whether the file is complete */
        filesQueued.insert(nzb_file.absoluteFilePath(), nzb_file.lastModified());

        if (!fileReadyTimer.isActive()) {
            fileReadyTimer.start(1000);
        }
    }
}

void AutoImport::fileReadyTimer_timeout()
{
    QMapIterator<QString, QDateTime> i(filesQueued);
    while (i.hasNext()) {
        i.next();

        /* Was the last change done at least 3 seconds ago? */
        if (i.value().secsTo(QDateTime::currentDateTime()) > 3) {
            /* Try to import the file */
            doImport(i.key());
        }
    }

    if (filesQueued.isEmpty()) {
        fileReadyTimer.stop();
    }
}

void AutoImport::doImport(QString path)
{
    QFile file(path);
    if (!file.open(QIODevice::ReadOnly)) {
        qWarning("%s", tr("Tried to import file '%1', but this failed as the file could not be opened").arg(path).toUtf8().data());
        filesQueued.remove(path);
        return;
    }

    QByteArray contents;
    contents = file.readAll();
    if (!contents.contains("</nzb>")) {
        qWarning("%s", tr("Tried to import file '%1', but this failed as the file is incomplete").arg(path).toUtf8().data());
        filesQueued.remove(path);
        return;
    }

    DialogImportNZB d(glue);
    QFileInfo info(path);
    QString collection_name = glue->utilsSanitizeCollectionName(info.baseName());
    d.quickImport(collection_name, QString(contents), false);

    filesQueued.remove(path);

    QString msg = tr("The file '%1'\nwas successfully imported with the collection name '%2'").arg(path, collection_name);
    mainwindow->showNotification(tr("NZB File Imported"), msg);

    /* Remove or move the file if necessary */
    QNNTPGrabGlue::ConfigOpts opts = glue->configGetOpts();
    if (opts.move_file_after_auto_import) {
        /* Create a folder named 'Imported' and move the file there */
         QDir d;
         QFileInfo info(file);
         d.mkdir(info.canonicalPath() + "/Imported");
         file.rename(info.canonicalPath() + "/Imported/" + info.fileName());
    } else {
        /* Remove the file */
        file.remove();
    }
}

void AutoImport::onPluginEvent(QString plugin_name, QString event_name, QList<QString> args)
{
    if (plugin_name != "Auto-import" || event_name != "nzb_imported") {
        return;
    }

    QString msg = tr("The file '%1'\non the NNTPGrab Server was successfully\nimported with the collection name '%2'").arg(args[0], args[1]);
    mainwindow->showNotification(tr("NZB File Imported"), msg);
}
