/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <glib.h>
#include <glib/gstdio.h>
#include "nntpgrab.h"
#include "nntpgrab_internal.h"
#include "nntpgrab_plugin.h"
#include "nntpgrab_utils.h"
#include "nntpconnection.h"
#include "queue_rawfile.h"
#include "download_queue.h"
#include "collection_alloc.h"
#include "configuration.h"
#include "decoder_thread.h"
#include "download_thread.h"
#include "plugins.h"

static GStaticMutex  mutex = G_STATIC_MUTEX_INIT;
static GList        *download_queue = NULL;
static int           num_servers = 0;
static int           num_enabled_servers = 0;
static unsigned int  is_high_priority_server;  /* These are actually a bit-masked variable where every server has one bit */
static unsigned int  is_normal_priority_server;
static unsigned int  is_server_enabled;
static NGConfigOpts  opts;

typedef struct _download_queue_server_position
{
    ngboolean task_probably_ready;

    int collection_pos;
    int file_pos;
    int part_pos;

    GList *collection_list;
    GList *file_list;
    GList *part_list;
} DownloadQueueServerPosition;

static DownloadQueueServerPosition server_positions[MAX_NNTP_SERVERS];

static void
rebuild_download_queue_list(NNTPCollection *collection)
{
    GList *list_file;

    // Clear out any old items in the list
    g_list_free(collection->files_to_download);
    collection->files_to_download = NULL;

    list_file = collection->files;
    while (list_file) {
        NNTPFile *file = list_file->data;

        if (!file->file_is_downloaded) {
            collection->files_to_download = g_list_append(collection->files_to_download, file);
        }

        list_file = g_list_next(list_file);
    }
}

static gboolean
get_bit_value(unsigned int param, int pos)
{
    unsigned int shifted = param >> pos;

    return ((shifted & 0x1) == 0x1);
}

static void
set_bit_value(unsigned int *param, int pos, gboolean value)
{
    unsigned int mask = 0x1 << pos;

    if (value) {
        *param = *param ^ mask;
    } else {
        mask = ~mask;
        *param = *param & mask;
    }
}

static gboolean
test_server_may_download_part(NNTPPart *part, int server_id)
{
    if (!get_bit_value(is_server_enabled, server_id)) {
        return FALSE;
    }

    if (get_bit_value(part->servers_already_tried, server_id)) {
        return FALSE;
    }

    // Is this server one with the highest priority ?
    if (get_bit_value(is_high_priority_server, server_id)) {
        return TRUE;
    }

    // Are all the other high-priority servers already tried ?
    if ((is_high_priority_server & part->servers_already_tried) != is_high_priority_server) {
        return FALSE;
    }

    // All the high-priority servers are already tried, now test against the servers with the 'normal' priority
    if (get_bit_value(is_normal_priority_server, server_id)) {
        return TRUE;
    }

    if ((is_normal_priority_server & part->servers_already_tried) != is_normal_priority_server) {
        return FALSE;
    }

    // All the normal priority servers are also tried, allow the part

    return TRUE;
}

static void
reset_server_positions(int server_id)
{
    server_positions[server_id].collection_pos = 0;
    server_positions[server_id].file_pos = 0;
    server_positions[server_id].part_pos = 0;

    server_positions[server_id].collection_list = NULL;
    server_positions[server_id].file_list = NULL;
    server_positions[server_id].part_list = NULL;

    server_positions[server_id].task_probably_ready = TRUE;
}

void
download_queue_reset_all_server_positions(void)
{
    int i;
    for (i = 0; i < MAX_NNTP_SERVERS; i++) {
        reset_server_positions(i);
    }
}

static void
check_and_update_server_position_if_necessary(int server_id, int collection_pos, int file_pos, int part_pos)
{
    /* Check if the active task is before or after the given parameters */
    if (collection_pos < server_positions[server_id].collection_pos) {
        /* Different collection, reset the cache completely */
        reset_server_positions(server_id);
    } else if (file_pos < server_positions[server_id].file_pos) {
        /* Different file, only reset the file and part positions */
        server_positions[server_id].file_list = NULL;
        server_positions[server_id].part_list = NULL;
        server_positions[server_id].file_pos = 0;
        server_positions[server_id].part_pos = 0;
    } else if (part_pos < server_positions[server_id].part_pos) {
        /* Same file, only reset the part position */
        server_positions[server_id].part_list = NULL;
        server_positions[server_id].part_pos = 0;
    }

    server_positions[server_id].task_probably_ready = TRUE;
}

static void
update_server_position_cache_if_necessary(NNTPCollection *collection, NNTPFile *file, NNTPPart *part, int server_id)
{
    int i;

    /* Find out the position of the given parameters */
    int collection_pos = g_list_index(download_queue, collection);
    int file_pos = g_list_index(collection->files, file);
    int part_pos = g_list_index(file->parts, part);

    /* Find out what the priority of the current server is */
    ngboolean high_prio_server = get_bit_value(is_high_priority_server, server_id);
    ngboolean normal_prio_server = get_bit_value(is_normal_priority_server, server_id);
    ngboolean low_prio_server = TRUE;

    /* Fix up the priority flags */
    if (high_prio_server) {
        normal_prio_server = TRUE;
        low_prio_server = TRUE;
    } else if (normal_prio_server) {
        low_prio_server = TRUE;
    }

    /* Update the position cache for all servers with an equal or lower priority */
    if (high_prio_server) {
        for (i = 0; i < MAX_NNTP_SERVERS; i++) {
            if (i != server_id && get_bit_value(is_high_priority_server, i)) {
                check_and_update_server_position_if_necessary(i, collection_pos, file_pos, part_pos);
            }
        }
    }

    if (normal_prio_server) {
        for (i = 0; i < MAX_NNTP_SERVERS; i++) {
            if (i != server_id && get_bit_value(is_normal_priority_server, i)) {
                check_and_update_server_position_if_necessary(i, collection_pos, file_pos, part_pos);
            }
        }
    }

    if (low_prio_server) {
        for (i = 0; i < MAX_NNTP_SERVERS; i++) {
            if (i != server_id && !get_bit_value(is_high_priority_server, i) && !get_bit_value(is_normal_priority_server, i)) {
                check_and_update_server_position_if_necessary(i, collection_pos, file_pos, part_pos);
            }
        }
    }
}

static void
start_to_decode_if_necessary(NNTPCollection *collection, NNTPFile *file)
{
    /* Check if all parts belonging to this part are downloaded so we can start decoding it */
    GList *list = file->parts;
    gboolean complete_flag = TRUE;
    while (list) {
        NNTPPart *part = list->data;

        if (part->now_downloading) {
            file->file_is_downloading = TRUE;
            complete_flag = FALSE;
        }

        if (!part->downloaded) {
            complete_flag = FALSE;
        }

        list = g_list_next(list);
    }

    /* All the parts are already downloaded? */
    if (complete_flag && !file->now_decoding && !file->file_is_downloaded) {
        DecoderData *decoder_data;

        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "File '%s' from collection '%s' is now ready to be decoded", file->subject, collection->collection_name);

        file->now_decoding = TRUE;

        nntpgrab_core_emit_file_state_changed(FALSE, collection->collection_name, file->subject, NULL, file->status, TASK_STATE_WAITING_FOR_DECODE, file->file_size_remaining, collection->total_size, collection->total_size_remaining);

        file->status = TASK_STATE_WAITING_FOR_DECODE;

        decoder_data = g_slice_new0(DecoderData);
        decoder_data->collection = collection;
        decoder_data->file = file;
        //decoder_data->poolPar2verify = poolPar2verify;
        //decoder_data->poolPar2repair = poolPar2repair;
        //decoder_data->poolUnpack = poolUnpack;

        file_ref(file);
        collection_ref(collection);

        decoder_thread_push_task(decoder_data);
    }
}

void
download_queue_check_if_files_are_ready_to_be_decoded(void)
{
    GList *list;

    g_static_mutex_lock(&mutex);

    list = download_queue;
    while (list) {
        NNTPCollection *collection = list->data;
        GList *list2;

        list2 = collection->files;
        while (list2) {
            NNTPFile *file = list2->data;

            start_to_decode_if_necessary(collection, file);

            list2 = g_list_next(list2);
        }

        list = g_list_next(list);
    }
    g_static_mutex_unlock(&mutex);
}

static gboolean
on_config_changed(Configuration *config, gpointer data)
{
    GList *list, *avail_servers;
    int i;

    g_static_mutex_lock(&mutex);

    opts = configuration_get_opts(config);

    is_high_priority_server = 0;
    is_normal_priority_server = 0;
    is_server_enabled = 0;
    num_enabled_servers = 0;
    num_servers = 0;

    avail_servers = configuration_get_avail_servers(config);
    list = avail_servers;
    i = 0;

    while (list) {
        char *servername = list->data;
        NGConfigServer *server = configuration_get_server_info(config, servername);

        g_return_val_if_fail(server != NULL, FALSE);

        if (server->enabled) {
            num_enabled_servers++;

            if (server->priority == SERVER_PRIORITY_HIGH) {
                set_bit_value(&is_high_priority_server, i, TRUE);
                set_bit_value(&is_normal_priority_server, i, FALSE);
            } else if (server->priority == SERVER_PRIORITY_LOW) {
                set_bit_value(&is_high_priority_server, i, FALSE);
                set_bit_value(&is_normal_priority_server, i, FALSE);
            } else {
                set_bit_value(&is_high_priority_server, i, FALSE);
                set_bit_value(&is_normal_priority_server, i, TRUE);;
            }
        }

        set_bit_value(&is_server_enabled, i, server->enabled);

        i++;
        num_servers++;

        g_free(servername);
        g_slice_free(NGConfigServer, server);

        list = g_list_next(list);
    }

    g_list_free(avail_servers);

    g_static_mutex_unlock(&mutex);

    return FALSE;
}

gboolean
download_queue_init(Configuration *config, char **errmsg)
{
    GList *list;
    int i;

    g_return_val_if_fail(download_queue == NULL, FALSE);

    if (!queue_rawfile_load(&download_queue, errmsg)) {
        /* Ignore read errors for now */
        g_free(*errmsg);
        *errmsg = NULL;
    }

    g_signal_connect(config, "config_changed", G_CALLBACK(on_config_changed), NULL);
    on_config_changed(config, NULL);

    list = download_queue;
    while (list) {
        NNTPCollection *collection = list->data;

        rebuild_download_queue_list(collection);

        list = g_list_next(list);
    }

    for (i = 0; i < MAX_NNTP_SERVERS; i++) {
        reset_server_positions(i);
    }

    return TRUE;
}

gboolean
download_queue_destroy(char **errmsg)
{
    GList *collections;

    g_static_mutex_lock(&mutex);

    if (!queue_rawfile_save(download_queue, errmsg)) {
        g_static_mutex_unlock(&mutex);
        return FALSE;
    }

    collections = download_queue;
    while (collections) {
        NNTPCollection *collection = (NNTPCollection *) collections->data;

        collection_unref(collection);

        collections = g_list_next(collections);
    }

    g_list_free(download_queue);
    download_queue = NULL;

    g_static_mutex_unlock(&mutex);

    return TRUE;
}

gboolean
download_queue_save(char **errmsg)
{
    gboolean retval;

    g_static_mutex_lock(&mutex);
    retval = queue_rawfile_save(download_queue, errmsg);
    g_static_mutex_unlock(&mutex);

    return retval;
}

gboolean
download_queue_get_next_part_to_download(NNTPCollection **collection, NNTPFile **file, NNTPPart **part, int server_id, gboolean *nothing_to_download_or_active)
{
    GList *list;

    g_return_val_if_fail(collection != NULL, FALSE);
    g_return_val_if_fail(file != NULL, FALSE);
    g_return_val_if_fail(part != NULL, FALSE);
    g_return_val_if_fail(server_id >= 0, FALSE);
    g_return_val_if_fail(nothing_to_download_or_active != NULL, FALSE);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "download_queue_get_next_part_to_download for server_id %i", server_id);

    *nothing_to_download_or_active = TRUE;

    if (num_enabled_servers == 0) {
        /* No enabled servers found, don't even bother checking the queue */
        return FALSE;
    }

    g_static_mutex_lock(&mutex);

    if (!server_positions[server_id].task_probably_ready) {
        /* It's uncertain whether there's anything left to download */
        /* Do only a quick scan of the now_decoding flag */
        goto check_decoded_state;
    }

    /* Do we already have a cached position for this server? */
    if (server_positions[server_id].collection_list) {
        list = server_positions[server_id].collection_list;
    } else {
        list = download_queue;
    }

    while (list) {
        GList *list2;

        *collection = (NNTPCollection *) list->data;

        // Is this collection being removed ?
        if ((*collection)->stop_flag) {
            list = g_list_next(list);
            continue;
        }

        /* Do we already have a cached position for this server? */
        if (server_positions[server_id].file_list) {
            list2 = server_positions[server_id].file_list;
        } else {
            list2 = (*collection)->files_to_download;
        }

        while (list2) {
            GList *list3;

            *file = (NNTPFile *) list2->data;

            /* Do we need to skip this file? */
            if ((*file)->status == TASK_STATE_SKIPPED) {
                /* Ignore */
                server_positions[server_id].part_list = NULL;
                list2 = g_list_next(list2);
                continue;
            }

            // Is this file already downloaded?
            if ((*file)->file_is_downloaded) {
                // Yes, remove it from the files_to_download list and start over
                (*collection)->files_to_download = g_list_remove((*collection)->files_to_download, list2->data);
                server_positions[server_id].part_list = NULL;
                list2 = (*collection)->files_to_download;
                continue;
            }

            if ((*file)->now_decoding) {
                server_positions[server_id].part_list = NULL;
                list2 = g_list_next(list2);
                continue;
            }

            // Is this file being removed?
            if ((*file)->stop_flag) {
                server_positions[server_id].part_list = NULL;
                list2 = g_list_next(list2);
                continue;
            }

            /* Do we already have a cached position for this server? */
            if (server_positions[server_id].part_list) {
                list3 = server_positions[server_id].part_list;
            } else {
                list3 = (*file)->parts;
            }

            while (list3) {
                *part = (NNTPPart *) list3->data;

                if ((*part)->now_downloading) {
                    *nothing_to_download_or_active = FALSE;
                }

                if (!(*part)->downloaded) {
                    if (!(*part)->now_downloading && test_server_may_download_part(*part, server_id)) {
                        (*part)->now_downloading = TRUE;
                        (*file)->file_is_downloading = TRUE;

                        server_positions[server_id].collection_list = list;
                        server_positions[server_id].file_list = list2;
                        server_positions[server_id].part_list = list3;
                        server_positions[server_id].collection_pos = g_list_index(download_queue, *collection);
                        server_positions[server_id].file_pos = g_list_index((*collection)->files, *file);
                        server_positions[server_id].part_pos = g_list_index((*file)->parts, *part);
                        server_positions[server_id].task_probably_ready = TRUE;

                        g_static_mutex_unlock(&mutex);

                        *nothing_to_download_or_active = FALSE;

                        file_ref(*file);
                        collection_ref(*collection);
                        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "Returned part %i from file '%s' from collection '%s' for server_id %i", (*part)->part_num, (*file)->subject, (*collection)->collection_name, server_id);

                        return TRUE;
                    }

                    // Are all the servers already tried ?
                    if (!(is_server_enabled & (*part)->servers_already_tried) == is_server_enabled) {
                        *nothing_to_download_or_active = FALSE;
                    }
                }

                list3 = g_list_next(list3);
            }

            server_positions[server_id].part_list = NULL;

            list2 = g_list_next(list2);
        }

        server_positions[server_id].file_list = NULL;

        list = g_list_next(list);
    }

check_decoded_state:

    /* Make sure that there's nothing in the 'waiting for decode' state */
    list = download_queue;
    while (list) {
        NNTPCollection *collection_tmp = (NNTPCollection *) list->data;
        GList *list2;

        // Is this collection being removed ?
        if (collection_tmp->stop_flag) {
            list = g_list_next(list);
            continue;
        }

        list2 = collection_tmp->files_to_download;
        while (list2) {
            NNTPFile *file_tmp = (NNTPFile *) list2->data;

            // Is this file being removed?
            if (file_tmp->stop_flag) {
                list2 = g_list_next(list2);
                continue;
            }

            if (file_tmp->now_decoding || !file_tmp->file_is_downloaded) {
                *nothing_to_download_or_active = FALSE;
            }

            list2 = g_list_next(list2);
        }

        list = g_list_next(list);
    }

    *collection = NULL;
    *file = NULL;
    *part = NULL;

    reset_server_positions(server_id);
    server_positions[server_id].task_probably_ready = FALSE;

    g_static_mutex_unlock(&mutex);

    return FALSE;
}

void
download_queue_update_part_status(const char *servername, int connection_id, NNTPCollection *collection, NNTPFile *file, NNTPPart *part, int server_id, gboolean success, gboolean reset_to_zero, gboolean dont_retry_anymore)
{
    g_return_if_fail(servername != NULL);
    g_return_if_fail(connection_id >= 0);
    g_return_if_fail(collection != NULL);
    g_return_if_fail(file != NULL);
    g_return_if_fail(part != NULL);
    g_return_if_fail(server_id >= 0);

    g_static_mutex_lock(&collection->mutex);

    /* Did the file just changed state? */
    if (file->status == TASK_STATE_SKIPPED) {
        /* Ignore the event */
        g_static_mutex_unlock(&collection->mutex);
        return;
    }

    part->now_downloading = FALSE;
    part->downloaded = success;
    file->file_is_downloading = FALSE;

    if (success) {
        nntpgrab_core_emit_part_done(FALSE, servername, connection_id, collection->collection_name, file->subject, part->part_num, part->size);

        /* Only update the administration when the file still exists in the download queue */
        if (!file->stop_flag) {
            if (file->file_size_remaining < part->size) {
                g_print(__FILE__ ":%i Invalid file size remaining!\n", __LINE__);
                g_print("file->file_size_remaining = %"G_GUINT64_FORMAT"\n", file->file_size_remaining);
                g_print("part->size = %i\n", part->size);
                file->file_size_remaining = part->size;
            }

            if (collection->total_size_remaining < part->size) {
                g_print(__FILE__ ":%i Invalid collection size remaining!\n", __LINE__);
                g_print("collection->total_size_remaining = %"G_GUINT64_FORMAT"\n", collection->total_size_remaining);
                g_print("part->size = %i\n", part->size);
                collection->total_size_remaining = part->size;
            }

            file->file_size_remaining -= part->size;
            collection->total_size_remaining -= part->size;
            file->num_parts_downloaded++;

            nntpgrab_core_emit_file_download_state_update(FALSE, collection->collection_name, file->subject, file->num_parts, file->num_parts_downloaded, file->num_parts_failed, file->file_size, file->file_size_remaining, collection->total_size, collection->total_size_remaining);
        }

        /* Make sure that we check for now files to be downloaded */
        server_positions[server_id].task_probably_ready = TRUE;
    } else if (!success && dont_retry_anymore) {
        set_bit_value(&part->servers_already_tried, server_id, TRUE);
    }

    if (!success) {
        /* Allow other servers to try this part again */
        update_server_position_cache_if_necessary(collection, file, part, server_id);

        if (!dont_retry_anymore) {
            /* Retry this part later on the same server. Flush the position cache */
            reset_server_positions(server_id);
        }
    }

    /* Are all the servers already tried ? */
    if (!success && dont_retry_anymore) {
        gboolean all_servers_tried = TRUE;

        if ((is_server_enabled & part->servers_already_tried) == is_server_enabled) {
            all_servers_tried = TRUE;
        } else {
            all_servers_tried = FALSE;
        }

        nntpgrab_core_emit_part_failed(FALSE, servername, connection_id, collection->collection_name, file->subject, part->part_num, part->size, all_servers_tried);

        if (all_servers_tried && !reset_to_zero) {
            if (file->file_size_remaining < part->size) {
                g_print("Invalid file size remaining!\n");
                g_print("file->file_size_remaining = %"G_GUINT64_FORMAT"\n", file->file_size_remaining);
                g_print("part->size = %i\n", part->size);
                file->file_size_remaining = part->size;
            }

            if (collection->total_size_remaining < part->size) {
                g_print("Invalid collection size remaining!\n");
                g_print("collection->total_size_remaining = %"G_GUINT64_FORMAT"\n", collection->total_size_remaining);
                g_print("part->size = %i\n", part->size);
                collection->total_size_remaining = part->size;
            }

            /* Assume the part was downloaded as all servers were tried and no one had the part we requested */
            part->downloaded = TRUE;

            file->num_parts_failed++;
            file->file_size_remaining -= part->size;
            collection->total_size_remaining -= part->size;

            if (!file->stop_flag) {
                nntpgrab_core_emit_file_download_state_update(FALSE, collection->collection_name, file->subject, file->num_parts, file->num_parts_downloaded, file->num_parts_failed, file->file_size, file->file_size_remaining, collection->total_size, collection->total_size_remaining);
            }
        }
    }

    if (reset_to_zero) {
        file->num_parts_downloaded = 0;
        file->num_parts_failed = 0;

        if (!file->stop_flag) {
            nntpgrab_core_emit_file_download_state_update(FALSE, collection->collection_name, file->subject, file->num_parts, file->num_parts_downloaded, file->num_parts_failed, file->file_size, file->file_size_remaining, collection->total_size, collection->total_size_remaining);
        }
    }

    start_to_decode_if_necessary(collection, file);

    g_static_mutex_unlock(&collection->mutex);
}

gboolean
download_queue_test_is_decoding_still_needed(NNTPCollection *collection, NNTPFile *file)
{
    g_return_val_if_fail(collection != NULL, FALSE);
    g_return_val_if_fail(file != NULL, FALSE);

    g_static_mutex_lock(&collection->mutex);
    if (file->stop_flag) {
        file->now_decoding = FALSE;
        g_static_mutex_unlock(&collection->mutex);

        file_unref(file);
        collection_unref(collection);

        return FALSE;
    }
    g_static_mutex_unlock(&collection->mutex);

    return TRUE;
}

void
download_queue_about_to_start_decoding(NNTPCollection *collection, NNTPFile *file)
{
    g_return_if_fail(collection != NULL);
    g_return_if_fail(file != NULL);

    nntpgrab_core_emit_file_state_changed(FALSE, collection->collection_name, file->subject, NULL, file->status, TASK_STATE_DECODING, file->file_size_remaining, collection->total_size, collection->total_size_remaining);

    g_static_mutex_lock(&collection->mutex);
    file->status = TASK_STATE_DECODING;
    g_static_mutex_unlock(&collection->mutex);
}

void
download_queue_mark_file_as_decoded_and_unref(NNTPCollection *collection, NNTPFile *file, const char *real_filename, NGTaskState status)
{
    g_return_if_fail(collection != NULL);
    g_return_if_fail(file != NULL);
    /* real_filename can be NULL */

    g_static_mutex_lock(&collection->mutex);

    /* Was the task restarted while the decoding was busy? */
    if (!file->now_decoding) {
        /* Yes it was, forget the status update */
        goto out;
    }

    if (status == TASK_STATE_FINISHED_COMPLETE || status == TASK_STATE_FINISHED_INCOMPLETE) {
        g_return_if_fail(real_filename != NULL);

        /* Update the real filename */
        strncpy((char*) file->real_filename, real_filename, sizeof(file->real_filename) - 1);

        if (file->file_type == NNTP_FILE_TYPE_UNKNOWN) {
            /* Check the extension again. It might have changed now that the decoding is finished */
            file->file_type = nntpgrab_utils_get_file_type_of_filename(file->real_filename);
        }

        file->file_is_downloaded = TRUE;
    } else if (status == TASK_STATE_FINISHED_NO_PARTS_AVAIL) {
        file->file_is_downloaded = TRUE;
    } else if (status == TASK_STATE_WAITING_FOR_DECODE) {
        /* The decoder wants us to re-queue the file. Probably some temporary error occurs (like disk full) */

    }

    nntpgrab_core_emit_file_state_changed(FALSE, collection->collection_name, file->subject, file->real_filename, file->status, status, file->file_size_remaining, collection->total_size, collection->total_size_remaining);
    file->status = status;

    file->now_decoding = FALSE;

out:
    g_static_mutex_unlock(&collection->mutex);

    download_queue_test_is_collection_downloaded(collection);

    file_unref(file);
    collection_unref(collection);
}

static gboolean
test_is_subject_already_in_collection(NNTPCollection *collection, const char *subject)
{
    GList *list;

    list = collection->files;
    while (list) {
        NNTPFile *file = (NNTPFile *) list->data;

        if (!strcmp(file->subject, subject)) {
            return TRUE;
        }

        list = g_list_next(list);
    }

    return FALSE;
}

static gboolean
test_if_collection_contains_same_poster(NNTPCollection *collection, const char *poster)
{
    GList *list;

    list = collection->files;
    while (list) {
        NNTPFile *file = (NNTPFile *) list->data;

        if (strcmp(file->poster, poster)) {
            /* Different poster found */
            return FALSE;
        }

        list = g_list_next(list);
    }

    return TRUE;
}

static gboolean
download_queue_add_file_to_queue_internal(const char *collection_name, NNTPFile *file, char **errmsg)
{
    GList *list;
    NNTPCollection *collection;
    int i;

    g_static_mutex_lock(&mutex);

    // Find out if the collection_name already exists in the download queue
    list = download_queue;
    while (list) {
        collection = (NNTPCollection *) list->data;

        if (!strcmp(collection->collection_name, collection_name)) {
            // Collection found!

            // Does the subject already exists in this collection ?
            if (test_is_subject_already_in_collection(collection, file->subject)) {
                if (errmsg) {
                    *errmsg = g_strdup_printf(_("There already is a file with subject '%s' part of collection '%s'"), file->subject, collection_name);
                }

                g_static_mutex_unlock(&mutex);

                return FALSE;
            }

            // Add the file to this collection
            collection->files = g_list_append(collection->files, file);
            collection->files_to_download = g_list_append(collection->files_to_download, file);
            file->file_size_remaining = file->file_size;
            if (file->status != TASK_STATE_SKIPPED) {
                collection->total_size += file->file_size;
                collection->total_size_remaining += file->file_size;
            }

            file->position = g_list_index(collection->files, file) + 1;

            nntpgrab_core_emit_file_added(FALSE, collection->collection_name, file->subject, file->poster, file->stamp, file->file_size, collection->total_size, collection->total_size_remaining, file->status, file->num_parts, file->groups);

            // Is this the same poster as all the other files in this collection?
            if (test_if_collection_contains_same_poster(collection, file->poster)) {
                if (strcmp(collection->poster, file->poster)) {
                    // The poster wasn't set for this collection yet
                    strncpy(collection->poster, file->poster, sizeof(collection->poster) - 1);
                    nntpgrab_core_emit_collection_modified(FALSE, collection_name, collection->poster);
                }
            } else {
                // Some other file contained a different poster, reset to empty
                if (strcmp(collection->poster, file->poster)) {
                    memset(collection->poster, 0, sizeof(collection->poster));
                    nntpgrab_core_emit_collection_modified(FALSE, collection_name, collection->poster);
                }
            }

            g_static_mutex_unlock(&mutex);

            download_thread_awaken_if_needed();

            /* We can't rely on the contents of the server positions cache, so reset it */
            for (i = 0; i < MAX_NNTP_SERVERS; i++) {
                reset_server_positions(i);
            }

            return TRUE;
        }

        list = g_list_next(list);
    }

    // Collection doesn't exist yet, add a new one
    collection = collection_new();
    strncpy((char *) collection->collection_name, collection_name, sizeof(collection->collection_name));
    strncpy((char *) collection->poster, file->poster, sizeof(collection->poster));
    download_queue = g_list_append(download_queue, collection);

    nntpgrab_core_emit_collection_added(FALSE, collection_name, collection->poster);

    // Finally add the task to the collection
    collection->files = g_list_append(collection->files, file);
    collection->files_to_download = g_list_append(collection->files_to_download, file);
    if (file->status != TASK_STATE_SKIPPED) {
        collection->total_size = file->file_size;
        collection->total_size_remaining = file->file_size;
    } else {
        collection->total_size = 0;
        collection->total_size_remaining = 0;
    }
    file->position = 1;

    nntpgrab_core_emit_file_added(FALSE, collection->collection_name, file->subject, file->poster, file->stamp, file->file_size, collection->total_size, collection->total_size_remaining, file->status, file->num_parts, file->groups);

    g_static_mutex_unlock(&mutex);

    download_thread_awaken_if_needed();

    return TRUE;
}

gboolean
download_queue_add_file_to_queue(const char *collection_name, const char *subject, const char *poster, time_t stamp, nguint64 file_size, NGList *groups, NGList *parts, char **errmsg)
{
    gboolean ret;
    NNTPFile *file;
    GList *list;
#ifndef DOWNLOAD_QUEUE_DONT_SET_TMP_FILENAME
    char *tmp;
    char *tmp2;
#endif
    GHashTable *hashtable;
    char collection_name_copy[256];          /* This parameter needs to be of the same size as defined in base/collections.h! */

    g_return_val_if_fail(collection_name != NULL, FALSE);
    g_return_val_if_fail(subject != NULL, FALSE);
    g_return_val_if_fail(poster != NULL, FALSE);
    g_return_val_if_fail(file_size >= 0, FALSE);
    g_return_val_if_fail(groups != NULL, FALSE);
    g_return_val_if_fail(parts != NULL, FALSE);

    memset(&collection_name_copy, 0, sizeof(collection_name_copy));
    strncpy(collection_name_copy, collection_name, sizeof(collection_name_copy) - 1);

    file = file_new();
    strncpy((char *) file->subject, subject, sizeof(file->subject));
    strncpy((char *) file->poster, poster, sizeof(file->poster));
    file->stamp = stamp;
    file->file_size = file_size;
    file->file_size_remaining = file_size;

#ifndef DOWNLOAD_QUEUE_DONT_SET_TMP_FILENAME
    tmp = tempnam(opts.temp_directory, NULL);
    tmp2 = g_path_get_basename(tmp);
    strncpy((char *) file->tmp_filename, tmp2, sizeof(file->tmp_filename));
    g_free(tmp2);
    free(tmp);                  // Don't use g_free here as tempnam is a libc function!
#endif

    // Detect the file type and the par2 blocks
    if (!nntpgrab_utils_strip_subject(file->subject, NULL, NULL, NULL, NULL, NULL, &file->file_type, &file->par2_start_num, &file->par2_num_blocks, NULL, &file->num_parts)) {
        // Sanitize values
        file->file_type = NNTP_FILE_TYPE_UNKNOWN;
        file->par2_start_num = -1;
        file->par2_num_blocks =-1;
        file->num_parts = g_list_length((GList*) parts);
    }

    /* PAR2 files need to be skipped by default when the intelligent PAR2 downloading feature is enabled */
    if (opts.enable_intelligent_par2_downloading) {
        if (file->file_type == NNTP_FILE_TYPE_PAR2) {
            file->status = TASK_STATE_SKIPPED;
        } else {
            file->status = TASK_STATE_WAITING_FOR_DOWNLOAD;
        }
    } else {
        file->status = TASK_STATE_WAITING_FOR_DOWNLOAD;
    }

    if (file->par2_start_num == -1) {
        file->par2_end_num = -1;
    } else {
        file->par2_end_num = file->par2_start_num + file->par2_num_blocks;
    }

    list = (GList*) groups;
    while (list) {
        const char *group = list->data;
        file->groups = g_list_append(file->groups, g_strdup(group));
        list = g_list_next(list);
    }

    // When adding parts, prevent the adding of duplicate part numbers
    hashtable = g_hash_table_new(g_direct_hash, g_direct_equal);

    list = (GList*) parts;
    while (list) {
        NNTPGrabPart *part = list->data;
        NNTPPart *new_part;

        new_part = g_hash_table_lookup(hashtable, GINT_TO_POINTER(part->part_num));
        if (new_part) {
            // Hey, we've seen this part before!
            // Only keep the newest only
            memset(new_part, 0, sizeof(NNTPPart));
        } else {
            // We've never seen this part before
            new_part = g_slice_new0(NNTPPart);
            file->parts = g_list_append(file->parts, new_part);
            g_hash_table_insert(hashtable, GINT_TO_POINTER(part->part_num), new_part);
        }

        strncpy((char *) new_part->message_id, part->message_id, sizeof(new_part->message_id));
        new_part->size = part->size;
        new_part->part_num = part->part_num;

        list = g_list_next(list);
    }

    g_hash_table_destroy(hashtable);

    ret = download_queue_add_file_to_queue_internal(collection_name_copy, file, errmsg);
    if (!ret) {
        file_unref(file);
    }

    return ret;
}

gboolean
download_queue_del_file_from_queue(const char *collection_name, const char *subject, char **errmsg)
{
    GList *list;
    NNTPCollection *collection;
    int i;
#if 0
    GKeyFile *keyfile;
#endif
    char collection_name_copy[256];          /* This parameter needs to be of the same size as defined in base/collections.h! */

    g_return_val_if_fail(collection_name != NULL, FALSE);
    /* subject can be NULL */

    memset(&collection_name_copy, 0, sizeof(collection_name_copy));
    strncpy(collection_name_copy, collection_name, sizeof(collection_name_copy) - 1);

    g_static_mutex_lock(&mutex);

#if 0
    if (!(keyfile = queue_gkeyfile_open(errmsg))) {
        g_static_mutex_unlock(&mutex);
        return FALSE;
    }

    if (!queue_gkeyfile_begin_transaction(keyfile, errmsg)) {
        g_static_mutex_unlock(&mutex);
        queue_gkeyfile_close(keyfile);
        return FALSE;
    }
#endif

    // Find out if the collection_name exists in the download queue
    list = download_queue;
    while (list) {
        collection = (NNTPCollection *) list->data;

        if (!strcmp(collection->collection_name, collection_name_copy)) {
            // Collection found!
            GList *list2;
            gboolean delete_flag = FALSE;

            // Do we have a subject?
            if (!subject) {
                // Delete the whole collection
                delete_flag = TRUE;

                // Mark the whole collection as being removed
                collection->stop_flag = TRUE;
            }

            // Try to find the subject
            list2 = collection->files;
            while (list2) {
                NNTPFile *file = (NNTPFile *) list2->data;

                list2 = g_list_next(list2);

                if (delete_flag || !strcmp(file->subject, subject)) {
                    // Bingo!
                    // Is this file being downloaded ?
                    while (file->file_is_downloading || file->now_decoding) {
                        // Mark this file as being removed
                        file->stop_flag = TRUE;

#if 0
                        // Release the lock and wait for the download to complete
                        g_static_mutex_unlock(&mutex);

                        g_usleep(G_USEC_PER_SEC / 10);

                        g_static_mutex_lock(&mutex);
#endif

                        // Make sure that when all parts are downloaded, but the decoder isn't triggered, we reset the flag
                        if (file->file_is_downloading && !file->now_decoding) {
                            GList *list3 = file->parts;
                            gboolean is_downloading_flag = FALSE;

                            while (list3) {
                                NNTPPart *part = (NNTPPart*) list3->data;

                                if (part->now_downloading) {
                                    is_downloading_flag = TRUE;
                                }

                                list3 = g_list_next(list3);
                            }

                            file->file_is_downloading = is_downloading_flag;
                        }

#if 1
                        break;
#endif
                    }

#if 0
                    // Remove the entry from the database
                    if (!queue_gkeyfile_remove_file_from_db(keyfile, collection, file, TRUE, errmsg)) {
                        g_static_mutex_unlock(&mutex);
                        queue_gkeyfile_close(keyfile);
                        return FALSE;
                    }
#endif

                    // Unlink any possible downloaded parts
                    for (i = 0; i < file->num_parts; i++) {
                        char *filename = g_strdup_printf("%s%s%s.%i", opts.temp_directory, G_DIR_SEPARATOR_S, file->tmp_filename, i + 1);
                        g_unlink(filename);
                        g_free(filename);
                    }

                    // Update the file size
                    if (file->status != TASK_STATE_SKIPPED) {
                        collection->total_size += file->file_size;
                        if (collection->total_size_remaining < file->file_size_remaining) {
                            g_print(__FILE__ ":%i invalid values for size_remaining: collection->collection_size_remaining = %"G_GINT64_FORMAT", file->file_size_remaining = %"G_GINT64_FORMAT"\n", __LINE__, collection->total_size_remaining, file->file_size_remaining);
                            collection->total_size_remaining = 0;
                        } else {
                            collection->total_size_remaining -= file->file_size_remaining;
                        }
                    }

                    // Remove it from the list
                    collection->files = g_list_remove(collection->files, file);
                    collection->files_to_download = g_list_remove(collection->files_to_download, file);

                    nntpgrab_core_emit_file_removed(FALSE, collection_name_copy, file->subject, collection->total_size, collection->total_size_remaining);

                    file_unref(file);

                    // Does the collection still got any items in it ?
                    if (g_list_length(collection->files) == 0) {
#if 0
                        // Yes, remove the entire collection
                        if (!queue_gkeyfile_remove_collection_from_db(keyfile, collection, errmsg)) {
                            g_static_mutex_unlock(&mutex);
                            queue_gkeyfile_close(keyfile);
                            return FALSE;
                        }
#endif

                        download_queue = g_list_remove(download_queue, collection);

                        nntpgrab_core_emit_collection_removed(FALSE, collection_name_copy);

                        collection_unref(collection);

                        g_static_mutex_unlock(&mutex);

#if 0
                        if (!queue_gkeyfile_end_transaction(keyfile, errmsg)) {
                            queue_gkeyfile_close(keyfile);
                            return FALSE;
                        }

                        queue_gkeyfile_close(keyfile);
#endif

                        /* We can't rely on the contents of the server positions cache, so reset it */
                        for (i = 0; i < MAX_NNTP_SERVERS; i++) {
                            reset_server_positions(i);
                        }

                        return TRUE;
                    }

                    if (!delete_flag) {
                        // Is this the same poster as all the other files in this collection?
                        file = collection->files->data;
                        if (test_if_collection_contains_same_poster(collection, file->poster)) {
                            if (strcmp(collection->poster, file->poster)) {
                                // The poster wasn't set for this collection yet
                                strncpy(collection->poster, file->poster, sizeof(collection->poster) - 1);
                                nntpgrab_core_emit_collection_modified(FALSE, collection_name_copy, collection->poster);
                            }
                        } else {
                            // Some other file contained a different poster, reset to empty
                            if (strcmp(collection->poster, file->poster)) {
                                memset(collection->poster, 0, sizeof(collection->poster));
                                nntpgrab_core_emit_collection_modified(FALSE, collection_name_copy, collection->poster);
                            }
                        }

#if 0
                        if (!queue_gkeyfile_end_transaction(keyfile, errmsg)) {
                            g_static_mutex_unlock(&mutex);
                            queue_gkeyfile_close(keyfile);
                            return FALSE;
                        }

                        queue_gkeyfile_close(keyfile);
#endif

                        /* We can't rely on the contents of the server positions cache, so reset it */
                        for (i = 0; i < MAX_NNTP_SERVERS; i++) {
                            reset_server_positions(i);
                        }

                        g_static_mutex_unlock(&mutex);
                        return TRUE;
                    }
                }
            }

            if (delete_flag) {
                g_static_mutex_unlock(&mutex);

#if 0
                if (!queue_gkeyfile_end_transaction(keyfile, errmsg)) {
                    return FALSE;
                }

                queue_gkeyfile_close(keyfile);
#endif

                /* We can't rely on the contents of the server positions cache, so reset it */
                for (i = 0; i < MAX_NNTP_SERVERS; i++) {
                    reset_server_positions(i);
                }

                return TRUE;
            }
        }

        list = g_list_next(list);
    }

    if (errmsg) {
        *errmsg = g_strdup_printf(_("Subject '%s' was not found in collection '%s'"), subject, collection_name_copy);
    }

    g_static_mutex_unlock(&mutex);

#if 0
    queue_gkeyfile_close(keyfile);
#endif

    return FALSE;
}

gboolean
download_queue_restart_file(const char *collection_name, const char *subject, char **errmsg)
{
    GList *list;
    int pos_last_file_to_download;
    NNTPCollection *collection;
    char collection_name_copy[256];          /* This parameter needs to be of the same size as defined in base/collections.h! */
    int i;

    g_return_val_if_fail(collection_name != NULL, FALSE);
    /* subject can be NULL */

    memset(&collection_name_copy, 0, sizeof(collection_name_copy));
    strncpy(collection_name_copy, collection_name, sizeof(collection_name_copy) - 1);

    g_static_mutex_lock(&mutex);

    // Find out if the collection_name already exists in the download queue
    list = download_queue;
    while (list) {
        collection = (NNTPCollection *) list->data;

        pos_last_file_to_download = 0;

        if (!strcmp(collection->collection_name, collection_name_copy)) {
            // Collection found!
            GList *list2;
            gboolean restart_flag = FALSE;

            // Do we have a subject?
            if (!subject) {
                // Restart the whole collection
                restart_flag = TRUE;

                // Clear out the old list with items which still need to be downloaded
                g_list_free(collection->files_to_download);
                collection->files_to_download = NULL;
            }

            list2 = collection->files;
            while (list2) {
                int pos;
                NNTPFile *file = list2->data;

                if (restart_flag || !strcmp(file->subject, subject)) {
                    // Subject found!
                    GList *list3;
                    gboolean file_is_downloading = FALSE;
                    int filesize_remaining_active = 0;

                    list3 = file->parts;
                    while (list3) {
                        NNTPPart *part = list3->data;

                        if (part->now_downloading) {
                            list3 = g_list_next(list3);
                            file_is_downloading = TRUE;
                            filesize_remaining_active += part->size;
                            continue;
                        }

                        part->downloaded = FALSE;
                        part->now_downloading = FALSE;
                        part->servers_already_tried = 0;

                        list3 = g_list_next(list3);
                    }

                    file->file_is_downloading = file_is_downloading;
                    file->file_is_downloaded = FALSE;
                    file->now_decoding = FALSE;
                    file->num_parts_downloaded = 0;
                    file->num_parts_failed = 0;

                    if (restart_flag) {
                        // If the whole collection is restarted, don't bother with merging at the right position in the files_to_download list
                        collection->files_to_download = g_list_append(collection->files_to_download, file);
                    } else if (!file_is_downloading) {
                        // Only a part of the collection has been restarted
                        // The position of the previous item which still needs to be downloaded
                        // has already been saved in list_last_file_to_download
                        if (pos_last_file_to_download > 0) {
                            pos_last_file_to_download++;
                        }
                        collection->files_to_download = g_list_insert(collection->files_to_download, file,  pos_last_file_to_download);
                    }

                    collection->total_size_remaining += file->file_size - file->file_size_remaining - filesize_remaining_active;
                    file->file_size_remaining = file->file_size - filesize_remaining_active;

                    nntpgrab_core_emit_file_state_changed(FALSE, collection_name_copy, file->subject, NULL, file->status, TASK_STATE_WAITING_FOR_DOWNLOAD, file->file_size_remaining, collection->total_size, collection->total_size_remaining);
                    file->status = TASK_STATE_WAITING_FOR_DOWNLOAD;

                    if (!restart_flag) {
                        g_static_mutex_unlock(&mutex);

                        download_thread_awaken_if_needed();

                        /* We can't rely on the contents of the server positions cache, so reset it */
                        for (i = 0; i < MAX_NNTP_SERVERS; i++) {
                            reset_server_positions(i);
                        }

                        return TRUE;
                    }
                }

                pos = g_list_index(collection->files_to_download, file);
                if (pos > 0) {
                    pos_last_file_to_download = pos;
                }

                list2 = g_list_next(list2);
            }

            if (restart_flag) {
                g_static_mutex_unlock(&mutex);

                download_thread_awaken_if_needed();

                /* We can't rely on the contents of the server positions cache, so reset it */
                for (i = 0; i < MAX_NNTP_SERVERS; i++) {
                    reset_server_positions(i);
                }

                return TRUE;
            }
        }

        list = g_list_next(list);
    }

    g_static_mutex_unlock(&mutex);

    if (errmsg) {
        *errmsg = g_strdup_printf(_("Subject '%s' was not found in collection '%s'"), subject, collection_name_copy);
    }

    return FALSE;
}

struct _foreach_data
{
    ForeachCollectionFunc collection_func;
    ForeachFileFunc file_func;
    ForeachGroupFunc group_func;
    gpointer data;
};

static void
foreach_func(gpointer data, gpointer user_data)
{
    struct _foreach_data *foreach_data = (struct _foreach_data *) user_data;
    NNTPCollection *collection = (NNTPCollection *) data;
    GList *list, *list2;

    if (foreach_data->collection_func) {
        foreach_data->collection_func(collection->collection_name, collection->poster, collection->total_size, collection->total_size_remaining, collection->position, foreach_data->data);
    }

    if (!foreach_data->group_func && !foreach_data->file_func) {
        return;
    }

    list = collection->files;
    while (list) {
        NNTPFile *file = (NNTPFile *) list->data;

        if (foreach_data->file_func) {
            gint64 file_size_remaining;

            if (file->status == TASK_STATE_SKIPPED) {
                file_size_remaining = 0;
            } else {
                file_size_remaining = file->file_size_remaining;
            }

            foreach_data->file_func(collection->collection_name, file->subject, file->poster, (gint64) file->stamp, file->file_size, file_size_remaining, file->position, g_list_length(file->parts), file->num_parts_downloaded, file->num_parts_failed, file->status, file->real_filename, foreach_data->data);
        }

        if (foreach_data->group_func) {
            list2 = file->groups;
            while (list2) {
                const char *group = (const char *) list2->data;

                foreach_data->group_func(collection->collection_name, file->subject, group, foreach_data->data);

                list2 = g_list_next(list2);
            }
        }

        list = g_list_next(list);
    }
}

void
download_queue_foreach_file(ForeachCollectionFunc collection_func, ForeachFileFunc file_func, ForeachGroupFunc group_func, gpointer data)
{
    struct _foreach_data foreach_data;

    foreach_data.collection_func = collection_func;
    foreach_data.file_func = file_func;
    foreach_data.group_func = group_func;
    foreach_data.data = data;

    g_static_mutex_lock(&mutex);
    g_list_foreach(download_queue, foreach_func, &foreach_data);
    g_static_mutex_unlock(&mutex);
}

gboolean
download_queue_move_file(const char *collection_name_src, const char *subject_src, const char *collection_name_dest, int position_dest)
{
    GList *list;
    NNTPCollection *collection_src = NULL;
    NNTPCollection *collection_dest = NULL;
    NNTPFile *file = NULL;
    gboolean found = FALSE;
    int old_position = 0;
    int i;
    guint hash_collection_name_src;
    guint hash_subject_src;

    g_return_val_if_fail(collection_name_src != NULL, FALSE);
    g_return_val_if_fail(subject_src != NULL, FALSE);
    g_return_val_if_fail(collection_name_dest != NULL, FALSE);
    g_return_val_if_fail(position_dest >= -1, FALSE);

    // Moving between collections isn't possible yet in version 0.3
    g_return_val_if_fail(!strcmp(collection_name_src, collection_name_dest), FALSE);

    hash_collection_name_src = g_str_hash(collection_name_src);
    hash_subject_src = g_str_hash(subject_src);

    g_static_mutex_lock(&mutex);

    // Find the subject
    list = download_queue;
    while (list && !found) {
        collection_src = (NNTPCollection *) list->data;

        if (g_str_hash(collection_src->collection_name) == hash_collection_name_src &&
            !strcmp(collection_src->collection_name, collection_name_src)) {

            GList *list2;

            list2 = collection_src->files;
            while (list2) {
                file = list2->data;

                if (g_str_hash(file->subject) == hash_subject_src &&
                    !strcmp(file->subject, subject_src)) {

                    found = TRUE;
                    break;
                }

                file = NULL;
                list2 = g_list_next(list2);
                old_position++;
            }
        }

        list = g_list_next(list);
    }

    if (!found) {
        g_static_mutex_unlock(&mutex);
        return FALSE;
    }

    // Is the collection_name_src the same as collection_name_dest ?
    if (!strcmp(collection_name_src, collection_name_dest)) {
        collection_dest = collection_src;
    } else {
        // Nope, we need to find the collection_dest
        list = download_queue;
        found = FALSE;

        while (list && !found) {
            collection_dest = (NNTPCollection *) list->data;

            if (g_str_hash(collection_dest->collection_name) == hash_collection_name_src &&
                !strcmp(collection_dest->collection_name, collection_name_src)) {

                found = TRUE;
            }

            list = g_list_next(list);
        }

        if (!found) {
            g_static_mutex_unlock(&mutex);
            return FALSE;
        }
    }

    // Perform the actual move
    collection_src->files = g_list_remove(collection_src->files, file);
    collection_dest->files = g_list_insert(collection_dest->files, file, position_dest);

    /* Walk through all the files in the collection and update the position if necessary */
    i = 0;
    list = collection_src->files;
    while (list) {
        NNTPFile *file_tmp = (NNTPFile *) list->data;

        if (file_tmp->position != i) {
            file_tmp->position = i;
        }

        i++;
        list = g_list_next(list);
    }

    if (collection_src != collection_dest) {
        /* If the files was moved to another collection, also update the other collection */
        i = 0;
        list = collection_dest->files;
        while (list) {
            NNTPFile *file_tmp = (NNTPFile *) list->data;

            if (file_tmp->position != i) {
                file_tmp->position = i;
            }

            i++;
            list = g_list_next(list);
        }
    }

    rebuild_download_queue_list(collection_src);
    rebuild_download_queue_list(collection_dest);

    /* Invalidate the server position cache */
    download_queue_reset_all_server_positions();

#if 0
    int i = 0;

    list = collection_dest->files;
    while (list) {
        NNTPFile *file = (NNTPFile *) list->data;

        g_print("subject %i = %s\n", i, file->subject);
        i++;
        list = g_list_next(list);
    }
#endif

    g_static_mutex_unlock(&mutex);

    nntpgrab_core_emit_file_moved(FALSE, collection_name_src, subject_src, collection_name_dest, old_position, position_dest);

    return TRUE;
}

gboolean
download_queue_move_collection(const char *collection_name, int new_position)
{
    GList *list;
    NNTPCollection *collection_src = NULL;
    gboolean found = FALSE;
    int orig_position = 0;
    guint hash_collection_name;

    g_return_val_if_fail(collection_name != NULL, FALSE);
    g_return_val_if_fail(new_position >= -1, FALSE);
    g_return_val_if_fail(new_position == -1 || new_position < g_list_length(download_queue), FALSE);

    hash_collection_name = g_str_hash(collection_name);

    g_static_mutex_lock(&mutex);

    // Find the subject
    list = download_queue;
    while (list) {
        collection_src = (NNTPCollection *) list->data;

        if (g_str_hash(collection_src->collection_name) == hash_collection_name &&
            !strcmp(collection_src->collection_name, collection_name)) {

            found = TRUE;
            download_queue = g_list_remove(download_queue, collection_src);
            download_queue = g_list_insert(download_queue, collection_src, new_position);

            break;
        }

        orig_position++;
        list = g_list_next(list);
    }

#if 0
    int i = 0;

    list = schedular->queue;
    while (list) {
        collection_src = (NNTPCollection *) list->data;

        g_print("collection %i = %s\n", i, collection_src->collection_name);
        i++;
        list = g_list_next(list);
    }
#endif

    g_static_mutex_unlock(&mutex);

    if (found) {
        /* Invalidate the server position cache */
        download_queue_reset_all_server_positions();

        nntpgrab_core_emit_collection_moved(FALSE, collection_name, orig_position, new_position);
    }

    return found;
}

gboolean
download_queue_mark_task_optional(const char *collection_name, const char *subject, gboolean is_optional)
{
    GList *list;
    NNTPCollection *collection;

    g_static_mutex_lock(&mutex);

    /* Find out if the collection_name exists in the download queue */
    list = download_queue;
    while (list) {
        collection = (NNTPCollection *) list->data;

        if (!strcmp(collection->collection_name, collection_name)) {
            /* Collection found! */
            /* Try to find the subject */
            GList *list2 = collection->files;
            while (list2) {
                NNTPFile *file = (NNTPFile *) list2->data;
                gboolean changed = FALSE;

                if (!strcmp(file->subject, subject)) {
                    if (is_optional && file->status == TASK_STATE_WAITING_FOR_DOWNLOAD) {
                        file->status = TASK_STATE_SKIPPED;

                        if (collection->total_size_remaining < file->file_size_remaining) {
                            g_print(__FILE__ ":%i Invalid file size remaining!\n", __LINE__);
                            g_print("collection->total_size_remaining = %"G_GUINT64_FORMAT"\n", collection->total_size_remaining);
                            g_print("file->file_size = %"G_GUINT64_FORMAT"\n", file->file_size);
                            g_print("file->file_size_remaining = %"G_GUINT64_FORMAT"\n", file->file_size_remaining);
                            collection->total_size_remaining = 0;
                        } else {
                            collection->total_size -= file->file_size;
                            collection->total_size_remaining -= file->file_size_remaining;
                        }

                        file->file_size_remaining = file->file_size;

                        nntpgrab_core_emit_file_state_changed(FALSE, collection_name, subject, file->real_filename, TASK_STATE_WAITING_FOR_DOWNLOAD, TASK_STATE_SKIPPED, file->file_size_remaining, collection->total_size, collection->total_size_remaining);

                        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, _("File '%s' of collection '%s' was marked optional"), collection_name, subject);

                        changed = TRUE;
                    } else if (!is_optional && file->status == TASK_STATE_SKIPPED) {
                        file->status = TASK_STATE_WAITING_FOR_DOWNLOAD;

                        collection->total_size += file->file_size;
                        collection->total_size_remaining += file->file_size;
                        file->file_size_remaining = file->file_size;

                        nntpgrab_core_emit_file_state_changed(FALSE, collection_name, subject, file->real_filename, TASK_STATE_SKIPPED, TASK_STATE_WAITING_FOR_DOWNLOAD, file->file_size_remaining, collection->total_size, collection->total_size_remaining);

                        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, _("File '%s' of collection '%s' was marked non-optional"), collection_name, subject);

                        changed = TRUE;
                    }

                    if (changed) {
                        rebuild_download_queue_list(collection);
                    }

                    g_static_mutex_unlock(&mutex);

                    download_thread_awaken_if_needed();

                    /* If the last file was marked as being non-optional, make sure that regular post-processing occurs */
                    if (changed) {
                        download_queue_test_is_collection_downloaded(collection);
                        download_queue_reset_all_server_positions();
                    }

                    return TRUE;
                }

                list2 = g_list_next(list2);
            }
        }

        list = g_list_next(list);
    }

    g_static_mutex_unlock(&mutex);

    /* File or collection not found */
    return FALSE;
}

typedef struct _par2_file_seen {
    char filename_without_extension[256];
    char filename[256];
    int  file_size;
    char subject[256];
} PAR2FileSeen;

void
download_queue_download_smallest_par2_files(NNTPCollection *collection)
{
    GList *list=  NULL;
    GList *list2 = NULL;
    GList *par2files = NULL;
    int i;

    /* Only download optional PAR2 files when the intelligent PAR2 downloading feature is enabled */
    if (!opts.enable_intelligent_par2_downloading) {
        return;
    }

    g_static_mutex_lock(&mutex);

    list = collection->files;
    while (list) {
        NNTPFile *file = (NNTPFile *) list->data;
        char *filename = NULL;
        char *extension;
        char filename_without_extension[256];
        gboolean filename_seen = FALSE;
        char *ptr;

        if (file->file_type != NNTP_FILE_TYPE_PAR2) {
            list = g_list_next(list);
            continue;
        }

        if (!nntpgrab_utils_strip_subject(file->subject, NULL, NULL, NULL, &filename, NULL, NULL, NULL, NULL, NULL, NULL)) {
            g_print(__FILE__ ":%i unable to strip subject: %s\n", __LINE__, file->subject);
            list = g_list_next(list);
            continue;
        }

        extension = g_strrstr(filename, ".");
        if (!extension) {
            /* File without extension, ignore */
            g_free(filename);
            continue;
        }

        if (g_ascii_strncasecmp(extension, ".par", 4) != 0) {
            /* Different extension, ignore */
            g_free(filename);
            continue;
        }

        memset(filename_without_extension, 0, sizeof(filename_without_extension));
        strncpy(filename_without_extension, filename, extension - filename);

        // Strip any .volXX+YY part
        ptr = g_strrstr(filename_without_extension, ".");
        if (ptr && !g_strncasecmp(ptr, ".vol", 4)) {
            *ptr = '\0';
        }

        list2 = par2files;
        while (list2) {
            PAR2FileSeen *par2file = (PAR2FileSeen*) list2->data;

            if (!strcmp(filename_without_extension, par2file->filename_without_extension)) {
                /* Is this the smallest PAR2 file? */
                if (file->file_size < par2file->file_size) {
                    /* Replace the old entry */
                    g_slice_free(PAR2FileSeen, par2file);
                    par2files = g_list_remove(par2files, par2file);
                    break;
                }

                /* We've already seen this PAR2 set, ignore */
                filename_seen = TRUE;
                break;
            }

            list2 = g_list_next(list2);
        }

        if (!filename_seen) {
            PAR2FileSeen *par2file_seen = g_slice_new0(PAR2FileSeen);

            strncpy(par2file_seen->filename_without_extension, filename_without_extension, sizeof(par2file_seen->filename_without_extension) - 1);
            strncpy(par2file_seen->filename, filename, sizeof(par2file_seen->filename) - 1);
            strncpy(par2file_seen->subject, file->subject, sizeof(par2file_seen->subject) - 1);
            par2file_seen->file_size = file->file_size;

            par2files = g_list_append(par2files, par2file_seen);
        }

        g_free(filename);

        list = g_list_next(list);
    }

    g_static_mutex_unlock(&mutex);

    /* The par2files array now contains a list of all smallest PAR2 files in the collection */
    list = par2files;
    while (list) {
        PAR2FileSeen *par2file_seen = (PAR2FileSeen*) list->data;

        download_queue_mark_task_optional(collection->collection_name, par2file_seen->subject, FALSE);
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Marked file '%s' as PAR2 index file", par2file_seen->subject);

        g_slice_free(PAR2FileSeen, par2file_seen);
        list = g_list_next(list);
    }
    g_list_free(par2files);

    /* We can't rely on the contents of the server positions cache, so reset it */
    for (i = 0; i < MAX_NNTP_SERVERS; i++) {
        reset_server_positions(i);
    }
}

void
download_queue_test_is_collection_downloaded(NNTPCollection *collection)
{
    gboolean collection_is_downloaded = TRUE;
    gboolean par2_files_detected = FALSE;
    gboolean par2_files_downloaded = FALSE;
    GList *list;

    /* Is everything downloaded? */
    g_static_mutex_lock(&mutex);

    list = collection->files;
    while (list) {
        NNTPFile *file = (NNTPFile*) list->data;

        if (file->file_type == NNTP_FILE_TYPE_PAR2) {
            par2_files_detected = TRUE;
        }
        if (file->file_type == NNTP_FILE_TYPE_PAR2 && file->status != TASK_STATE_SKIPPED) {
            par2_files_downloaded = TRUE;
        }

        if (file->status != TASK_STATE_SKIPPED && !file->file_is_downloaded) {
            collection_is_downloaded = FALSE;
            break;
        }

        list = g_list_next(list);
    }

    if (!collection_is_downloaded) {
        g_static_mutex_unlock(&mutex);
        return;
    }

    /* Check if we need to enable downloading of some PAR2 files */
    if (par2_files_detected && !par2_files_downloaded) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, _("Collection '%s' has been completely downloaded without the PAR2 files"), collection->collection_name);
        g_static_mutex_unlock(&mutex);
        download_queue_download_smallest_par2_files(collection);
        return;
    }

    /* All files downloaded, allow other plugins to do post-processing (like automatic PAR2 repair and unpack) */
    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, _("Collection '%s' has been completely downloaded"), collection->collection_name);
    ng_plugin_emit_internal_event("collection_downloaded", collection->collection_name);

    /* If neither auto par2 nor auto unpack are enabled, remove the collection automatically if necessary */
    if (!opts.enable_auto_unpack && !opts.enable_par2_repair && opts.auto_remove_collections_after_download) {
        char *errmsg = NULL;
        char *collection_name = g_strdup(collection->collection_name);

        g_static_mutex_unlock(&mutex);

        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, _("Now trying to remove collection '%s'"), collection_name);

        if (!download_queue_del_file_from_queue(collection_name, NULL, &errmsg)) {
            ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_WARNING, _("Removing collection '%s' failed: %s"), collection_name, errmsg);
            g_free(errmsg);
        }

        g_free(collection_name);
        g_static_mutex_lock(&mutex);
    }

    g_static_mutex_unlock(&mutex);

    download_queue_test_if_all_downloads_are_completed();
}

void
download_queue_test_if_all_downloads_are_completed(void)
{
    GList *list;

    g_static_mutex_lock(&mutex);

    list = download_queue;
    while (list) {
        GList *list2;
        NNTPCollection *collection = (NNTPCollection *) list->data;

        list2 = collection->files;
        while (list2) {
            NNTPFile *file = (NNTPFile *) list2->data;

            if (!file->file_is_downloaded && file->status != TASK_STATE_SKIPPED) {
                g_static_mutex_unlock(&mutex);
                return;
            }
            list2 = g_list_next(list2);
        }

        list = g_list_next(list);
    }

    g_static_mutex_unlock(&mutex);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, _("All downloads are completed"));

    /* FIXME: Keep track of PAR2 and Unpack actions */

    /* All files in all collections are downloaded */
    nntpgrab_core_emit_all_downloads_completed(FALSE);
}

void
download_queue_on_par2_repair_failure(NGPlugin *plugin_data, int num_params, const char **params, void *data)
{
    GList *list;
    const char *collection_name = params[0];
    char *active_par2_filename;
    int num_blocks_more_required = atoi(params[3]);
    int num_par2_files_enabled = 0;
    char par2set_name[128];

    /* Only do something when the intelligent PAR2 downloading feature is enabled */
    if (!opts.enable_intelligent_par2_downloading) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, _("%s:%i A PAR2 repair failure has occured but it's being ignored as the configuration option 'Only download PAR2 files when needed' is disabled"), __FILE__, __LINE__);
        return;
    }

    if (num_blocks_more_required == -1) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, _("%s:%i No PAR2 files will be enabled automatically as the par2 recovery failed due to an unknown error"), __FILE__, __LINE__);
        return;
    }

    active_par2_filename = g_path_get_basename(params[1]);

    memset(par2set_name, 0, sizeof(par2set_name));
    if (!nntpgrab_utils_extract_par2set_name_from_par2_filename(active_par2_filename, (char*) &par2set_name, sizeof(par2set_name))) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_WARNING, _("%s:%i Unable to extract the name of the PAR2 set belonging to file '%s'"), __FILE__, __LINE__, active_par2_filename);
        g_free(active_par2_filename);
        return;
    }

    /* Check if there are PAR2 items in the download queue which were skipped and can help to complete the repair */
    g_static_mutex_lock(&mutex);

    list = download_queue;
    while (list) {
        NNTPCollection *collection = (NNTPCollection*) list->data;
        GList *list2;

        /* Check if this is the right collection */
        if (strcmp(collection->collection_name, collection_name)) {
            list = g_list_next(list);
            continue;
        }

        list2 = collection->files;
        while (list2) {
            NNTPFile *file = (NNTPFile*) list2->data;
            char par2set_name2[128];
            char *filename = NULL;

            /* We're only interested in PAR2 files which aren't downloaded already */
            if (file->file_type != NNTP_FILE_TYPE_PAR2 ||
                file->status != TASK_STATE_SKIPPED) {

                list2 = g_list_next(list2);
                continue;
            }

            /* Extract the filename */
            if (!nntpgrab_utils_strip_subject(file->subject, NULL, NULL, NULL, &filename, NULL, NULL, NULL, NULL, NULL, NULL)) {
                ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, _("%s:%i Unable to extract filename from subject '%s'"), __FILE__, __LINE__, file->subject);
                list2 = g_list_next(list2);
                continue;
            }

            /* Extract the PAR2 set name */
            memset(par2set_name2, 0, sizeof(par2set_name2));
            if (!nntpgrab_utils_extract_par2set_name_from_par2_filename(filename, (char*) &par2set_name2, sizeof(par2set_name2))) {
                ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, _("%s:%i Unable to extract the name of the PAR2 set belonging to file '%s'"), __FILE__, __LINE__, filename);
                ng_free(filename);
                list2 = g_list_next(list2);
                continue;
            }

            /* Does this file belong to the same PAR2 set? */
            if (strstr(par2set_name2, par2set_name)) {        /* We use strstr instead of strcmp here as some posters forget to add quotes to the subject to mark the filename */
                /* Bingo! */
                g_static_mutex_unlock(&mutex);
                download_queue_mark_task_optional(collection->collection_name, file->subject, FALSE);
                g_static_mutex_lock(&mutex);

                num_par2_files_enabled++;
                num_blocks_more_required -= file->par2_num_blocks;
                if (num_blocks_more_required < 0) {
                    ng_free(filename);
                    break;
                }
            }

            ng_free(filename);
            list2 = g_list_next(list2);
        }

        /* When we've gotten here, the collection has been processed */
        break;
    }

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, _("%s:%i %i PAR2 files were enabled by the intelligent PAR2 downloader"), __FILE__, __LINE__, num_par2_files_enabled);

    g_static_mutex_unlock(&mutex);
    g_free(active_par2_filename);
}
