/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <glib-object.h>
#include <gmodule.h>
#include <errno.h>

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

#include "strptime.h"
#include "nntpconnection.h"
#include "nntpconnection_utils.h"
#include "nntpgrab_internal.h"
#include "nntpgrab_plugin.h"
#include "download_queue.h"
#include "download_thread.h"
#include "collection_alloc.h"

#ifndef WIN32
#define O_BINARY 0
#endif

/* declaration for throttle.c */
void throttle_pause(struct timeval start_time, off_t xferlen, int max_bandwidth);

#ifdef WIN32
void timersub(struct timeval *a, struct timeval *b, struct timeval *res)
{
    res->tv_sec = a->tv_sec - b->tv_sec;
    res->tv_usec = a->tv_usec - b->tv_usec;
    if (res->tv_usec < 0) {
        res->tv_sec--;
        res->tv_usec += 1000000;
    }
}
#endif

/*************************************/
/* Initialization and authentication */
/*************************************/

static void nntpconnection_login_username_sent(NntpConnection *conn);
static void nntpconnection_login_password_sent(NntpConnection *conn);
static void nntpconnection_login_set_mode_reader(NntpConnection *conn);
static void nntpconnection_login_process_mode_reader(NntpConnection *conn);
static void nntpconnection_process_body_command(NntpConnection *conn);
static void nntpconnection_process_body_data(NntpConnection *conn);
static void nntpconnection_process_group_command(NntpConnection *conn);
static void nntpconnection_send_xover_command(NntpConnection *conn);
static void nntpconnection_process_xover_command(NntpConnection *conn);
static void nntpconnection_process_xover_data(NntpConnection *conn);

void
nntpconnection_process_welcome_msg(NntpConnection *conn)
{
    char welcome_msg[4096];

    memset(welcome_msg, 0, sizeof(welcome_msg));

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(welcome_msg) - 1, welcome_msg, NULL, NULL)) {
        return;
    }

    nntpconnection_utils_strip_newline(welcome_msg, NULL);

    switch (nntpconnection_utils_get_status_code(welcome_msg)) {
        case 200:
        case 201:
            /* Welcome message is OK */
            nntpgrab_core_emit_connected(FALSE, conn->server_info.servername, conn->connection_id, welcome_msg);
            break;

        default:
            /* Welcome message is NOT ok, probably too many connections from this IP address */
            nntpconnection_disconnect_from_server(conn, DISCONNECT_TOO_MANY_CONNECTIONS, welcome_msg, __FILE__, __LINE__);

            return;
    }

    /* Do we need to send a username and password to the server? */
    if (strlen(conn->server_info.username) > 0) {
        char buf[1024];
        memset(buf, 0, sizeof(buf));

        snprintf(buf, sizeof(buf) - 1, "AUTHINFO USER %s\r\n", conn->server_info.username);
        if (!nntpconnection_send_msg(conn, buf, strlen(buf))) {
            return;
        }

        conn->active_method = nntpconnection_login_username_sent;
    } else {
        /* We don't need to log in. Prepare the connection for further communication */
        nntpconnection_login_set_mode_reader(conn);
    }
}

static void
nntpconnection_login_username_sent(NntpConnection *conn)
{
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, NULL)) {
        return;
    }

    if (nntpconnection_utils_get_status_code(buf) != 381) {
        /* Unexpected response code. Abort the connection */
        nntpconnection_utils_strip_newline(buf, NULL);
        nntpconnection_disconnect_from_server(conn, DISCONNECT_LOGIN_FAILURE, buf, __FILE__, __LINE__);
        return;
    }

    if (strlen(conn->server_info.password) > 0) {
        memset(buf, 0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AUTHINFO PASS %s\r\n", conn->server_info.password);
        if (!nntpconnection_send_msg(conn, buf, strlen(buf))) {
            return;
        }

        conn->active_method = nntpconnection_login_password_sent;
    } else {
        /* No password configured for this server. Try to prepare the connection for further communication */
        nntpconnection_login_set_mode_reader(conn);
    }
}

static void
nntpconnection_login_password_sent(NntpConnection *conn)
{
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, NULL)) {
        return;
    }

    if (nntpconnection_utils_get_status_code(buf) != 281) {
        /* Unexpected response code. Abort the connection */
        nntpconnection_utils_strip_newline(buf, NULL);
        nntpconnection_disconnect_from_server(conn, DISCONNECT_LOGIN_FAILURE, buf, __FILE__, __LINE__);
        return;
    }

    /* Login OK, prepare the connection for further communication */
    nntpconnection_login_set_mode_reader(conn);
}

static void
nntpconnection_login_set_mode_reader(NntpConnection *conn)
{
    if (!nntpconnection_send_msg(conn, "MODE READER\r\n", -1)) {
        return;
    }

    conn->active_method = nntpconnection_login_process_mode_reader;
}

static void
nntpconnection_login_process_mode_reader(NntpConnection *conn)
{
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, NULL)) {
        return;
    }

    switch (nntpconnection_utils_get_status_code(buf)) {
        case 200:
        case 201:
            /* Everything's ready! */
            break;

        case 480:
            /* Some usenet servers use this to notify there are too many connections active */
            nntpconnection_utils_strip_newline(buf, NULL);
            nntpconnection_disconnect_from_server(conn, DISCONNECT_TOO_MANY_CONNECTIONS, buf, __FILE__, __LINE__);
            return;

        default:
            /* Unexpected response code. Abort the connection */
            nntpconnection_utils_strip_newline(buf, NULL);
            nntpconnection_disconnect_from_server(conn, DISCONNECT_INVALID_MSG, buf, __FILE__, __LINE__);
            return;
    }

    /* The connection is now fully operational. Check if we need to download something */
    if (conn->job_type == NNTP_JOB_TYPE_ARTICLE) {
        if (conn->server_info.send_group_command) {
            nntpconnection_send_group_command(conn);
        } else {
            nntpconnection_send_body_command(conn);
        }
    } else if (conn->job_type == NNTP_JOB_TYPE_XOVER) {
        nntpconnection_send_xover_command(conn);
    } else {
        conn->active_method = NULL;
        conn->is_idle = TRUE;
        conn->idle_start_stamp = time(NULL);
    }
}

/**********************************************************/
/* body / article                                         */
/**********************************************************/
void
nntpconnection_send_body_command(NntpConnection *conn)
{
    char cmd[1024];

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->job_type == NNTP_JOB_TYPE_ARTICLE);
    g_return_if_fail(conn->collection != NULL);
    g_return_if_fail(conn->file != NULL);
    g_return_if_fail(conn->part != NULL);
    g_return_if_fail(conn->article_fd == -1);
    g_return_if_fail(strlen(conn->article_filename) > 0);

    nntpgrab_core_emit_part_download_start(FALSE, conn->server_info.servername, conn->connection_id, conn->collection->collection_name, conn->file->subject, conn->part->part_num);

    memset(cmd, 0, sizeof(cmd));
    snprintf(cmd, sizeof(cmd) - 1, "BODY %s\r\n", conn->part->message_id);
    if (!nntpconnection_send_msg(conn, cmd, strlen(cmd))) {
        return;
    }

    conn->active_method = nntpconnection_process_body_command;
    conn->article_bytes_downloaded = 0;
    conn->article_write_buffer_length = 0;
}

static void
nntpconnection_process_body_command(NntpConnection *conn)
{
    gboolean more_data_ready = FALSE;
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->job_type == NNTP_JOB_TYPE_ARTICLE);
    g_return_if_fail(conn->collection != NULL);
    g_return_if_fail(conn->file != NULL);
    g_return_if_fail(conn->part != NULL);
    g_return_if_fail(conn->article_fd == -1);
    g_return_if_fail(strlen(conn->article_filename) > 0);

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, &more_data_ready)) {
        return;
    }

    switch (nntpconnection_utils_get_status_code(buf)) {
        case 222:                       /* article retrieved - body follows */
            /* Open the file where we need to save our data to */
            conn->article_fd = open(conn->article_filename, O_CREAT | O_WRONLY | O_TRUNC | O_BINARY, S_IRUSR | S_IWUSR);
            if (conn->article_fd == -1) {
                download_thread_abort_without_waiting(_("Unable to create a file named '%s': %s"), conn->article_filename, strerror(errno));
                conn->active_method = NULL;
                return;
            }

            conn->active_method = nntpconnection_process_body_data;

            if (more_data_ready) {
                conn->active_method(conn);
            }

            return;

        case 430:                       /* no such article found */
            download_queue_update_part_status(conn->server_info.servername, conn->connection_id, conn->collection, conn->file, conn->part, conn->server_id, FALSE, FALSE, TRUE);

            collection_unref(conn->collection);
            file_unref(conn->file);
            conn->collection = NULL;
            conn->file = NULL;
            conn->part = NULL;
            conn->job_type = NNTP_JOB_TYPE_NONE;

            conn->active_method = NULL;
            conn->is_idle = TRUE;
            conn->idle_start_stamp = time(NULL);

            return;

        case 220:                       /* article retrieved - head and body follow */
        case 221:                       /* article retrieved - head follows */
        case 223:                       /* article retrieved - request text separately */
        case 412:                       /* no newsgroup has been selected */
        case 420:                       /* no current article has been selected */
        case 423:                       /* no such article number in this group */
        default:
            /* Unexpected response code. Abort the connection */
            nntpconnection_utils_strip_newline(buf, NULL);
            nntpconnection_disconnect_from_server(conn, DISCONNECT_INVALID_MSG, buf, __FILE__, __LINE__);
            return;
    }
}

static void
trim_double_dots(NntpConnection *conn, char *buf, int *length)
{
    g_return_if_fail(conn != NULL);
    g_return_if_fail(buf != NULL);
    g_return_if_fail(length != NULL);
    g_return_if_fail(*length > 0);

    if (*length >= 2 && buf[0] == '.' && buf[1] == '.') {
        memmove(buf, buf + 1, *length - 1);
        buf[*length - 1] = '\0';
        (*length)--;
    }
}

static void
nntpconnection_process_body_data(NntpConnection *conn)
{
    int length = 0;
    char buf[32768];
    gboolean more_data_ready = TRUE;
    struct timeval tv;
    struct timeval tv_diff;

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->job_type == NNTP_JOB_TYPE_ARTICLE);
    g_return_if_fail(conn->part != NULL);
    g_return_if_fail(conn->article_fd >= 0);

    while (more_data_ready) {
        memset(buf, 0, sizeof(buf));

        /* Keep on reading data until we've found the '\n.\r\n' or '\n.\n' sequence */

        if (!nntpconnection_read_msg(conn, FALSE, sizeof(buf) - 1, buf, &length, &more_data_ready)) {
            return;
        }

        g_return_if_fail(length > 0);

        conn->article_bytes_downloaded += length;
        if (conn->article_bytes_downloaded > conn->part->size) {
            conn->article_bytes_downloaded = conn->part->size;
        }

        gettimeofday(&tv, NULL);
        timersub(&tv, &conn->last_article_progress_announce, &tv_diff);
        if (tv_diff.tv_sec > 0 || tv_diff.tv_usec > G_USEC_PER_SEC / 10) {
            conn->last_article_progress_announce = tv;
            nntpgrab_core_emit_part_progress_update(FALSE, conn->server_info.servername, conn->connection_id, conn->collection->collection_name, conn->file->subject, conn->part->part_num, conn->article_bytes_downloaded, conn->part->size);
        }

        /* Check for the end sequence */
        if ((length == 3 && !strncmp(buf, ".\r\n", 3)) ||
            (length == 2 && !strncmp(buf, ".\n", 2))) {

            /* End sequence found! */

            /* Flush the buffers and close the file descriptor */
            if ((conn->article_write_buffer_length > 0 && write(conn->article_fd, conn->article_write_buffer, conn->article_write_buffer_length) != conn->article_write_buffer_length)) {
                /* Write error! Kill the download thread */
                download_thread_abort_without_waiting(_("%s:%i Unable to write article data to file: %s"), __FILE__, __LINE__, strerror(errno));

                /* Mark the part as failed so it will be retried later */
                ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Part %i of file '%s' from collection '%s' was completed successfully, but the writing to disk failed", conn->part->part_num, conn->collection->collection_name, conn->file->subject);
                download_queue_update_part_status(conn->server_info.servername, conn->connection_id, conn->collection, conn->file, conn->part, conn->server_id, FALSE, FALSE, FALSE);
            } else {
                /* Download and save succeeded */
                download_queue_update_part_status(conn->server_info.servername, conn->connection_id, conn->collection, conn->file, conn->part, conn->server_id, TRUE, FALSE, TRUE);
                ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "Part %i of file '%s' from collection '%s' was downloaded successfully", conn->part->part_num, conn->collection->collection_name, conn->file->subject);
            }

            close(conn->article_fd);
            conn->article_fd = -1;

            conn->article_bytes_downloaded = 0;
            conn->article_write_buffer_length = 0;

            memset(conn->article_write_buffer, 0, sizeof(conn->article_write_buffer));

            collection_unref(conn->collection);
            file_unref(conn->file);
            conn->collection = NULL;
            conn->file = NULL;
            conn->part = NULL;
            conn->job_type = NNTP_JOB_TYPE_NONE;
            conn->active_method = NULL;
            conn->is_idle = TRUE;

            conn->idle_start_stamp = time(NULL);

            return;
        }

        /* Un-escape any double-dots */
        trim_double_dots(conn, buf, &length);

        /* Keep the data in a temporary buffer to minimize the amount of I/O */
        if (conn->article_write_buffer_length + length > sizeof(conn->article_write_buffer)) {
            /* Flush the buffer */
            int pos = 0;

            do {
                int retval = write(conn->article_fd, conn->article_write_buffer + pos, conn->article_write_buffer_length - pos);
                if (retval == -1 || retval == 0) {
                    /* Write error! Kill the download thread */
                    char msg[1024];

                    snprintf(msg, sizeof(msg) - 1, _("%s:%i Unable to write article data to file: %s (errno = %i)"), __FILE__, __LINE__, strerror(errno), errno);
                    download_thread_abort_without_waiting("%s", msg);

                    /* Mark the part as failed so it will be retried later */
                    download_queue_update_part_status(conn->server_info.servername, conn->connection_id, conn->collection, conn->file, conn->part, conn->server_id, FALSE, FALSE, FALSE);

                    nntpconnection_disconnect_from_server(conn, DISCONNECT_WRITE_ERROR, msg, __FILE__, __LINE__);

                    return;
                } else {
                    pos += retval;
                }
            } while (pos != conn->article_write_buffer_length);

            memset(conn->article_write_buffer, 0, sizeof(conn->article_write_buffer));
            conn->article_write_buffer_length = 0;
        }

        memcpy(conn->article_write_buffer + conn->article_write_buffer_length, buf, length);
        conn->article_write_buffer_length += length;
    }
}

/**************************************************************
 * group / xover
 **************************************************************/
void
nntpconnection_send_group_command(NntpConnection *conn)
{
    char cmd[1024];

    g_return_if_fail(conn != NULL);
    g_return_if_fail(strlen(conn->newsgroup) > 0);

    memset(cmd, 0, sizeof(cmd));
    snprintf(cmd, sizeof(cmd) - 1, "GROUP %s\r\n", conn->newsgroup);
    if (!nntpconnection_send_msg(conn, cmd, strlen(cmd))) {
        return;
    }

    conn->active_method = nntpconnection_process_group_command;
}

static void
nntpconnection_process_group_command(NntpConnection *conn)
{
    gboolean more_data_ready = FALSE;
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    g_return_if_fail(conn != NULL);

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, &more_data_ready)) {
        return;
    }

    switch (nntpconnection_utils_get_status_code(buf)) {
        case 211:
            /*  211 n f l s group selected
                    (n = estimated number of articles in group,
                     f = first article number in the group,
                     l = last article number in the group,
                     s = name of the group.)
             */

            /* Send a BODY or XOVER command */
            if (conn->part) {
                nntpconnection_send_body_command(conn);
            } else {
                nntpconnection_send_xover_command(conn);
            }

            return;

        case 411:       /* No such group */
            nntpconnection_utils_strip_newline(buf, NULL);

            /* Only bail out when in XOVER mode */
            if (conn->part) {
                nntpconnection_send_body_command(conn);
            } else {
                nntpconnection_disconnect_from_server(conn, DISCONNECT_NORMAL, buf, __FILE__, __LINE__);
            }

            return;

        default:
            /* Unexpected response code. Abort the connection */
            nntpconnection_utils_strip_newline(buf, NULL);
            nntpconnection_disconnect_from_server(conn, DISCONNECT_INVALID_MSG, buf, __FILE__, __LINE__);
            return;
    }
}

void
nntpconnection_send_xover_command(NntpConnection *conn)
{
    char cmd[1024];

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->job_type == NNTP_JOB_TYPE_XOVER);
    g_return_if_fail(strlen(conn->newsgroup) > 0);
    g_return_if_fail(conn->xover_start_range >= -1);
    g_return_if_fail(conn->xover_end_range >= -1);

//    nntpgrab_core_emit_part_download_start(FALSE, conn->server_info.servername, conn->connection_id, conn->collection->collection_name, conn->file->subject, conn->part->part_num);

    memset(cmd, 0, sizeof(cmd));
    if (conn->xover_end_range == -1) {
        snprintf(cmd, sizeof(cmd) - 1, "XOVER %"G_GINT64_FORMAT"-\r\n", conn->xover_start_range);
    } else {
        snprintf(cmd, sizeof(cmd) - 1, "XOVER %"G_GINT64_FORMAT"-%"G_GINT64_FORMAT"\r\n", conn->xover_start_range, conn->xover_end_range);
    }

    if (!nntpconnection_send_msg(conn, cmd, strlen(cmd))) {
        return;
    }

    conn->active_method = nntpconnection_process_xover_command;
}

static void
nntpconnection_process_xover_command(NntpConnection *conn)
{
    gboolean more_data_ready = FALSE;
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->job_type == NNTP_JOB_TYPE_XOVER);

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, &more_data_ready)) {
        return;
    }

    switch (nntpconnection_utils_get_status_code(buf)) {
        case 224:                       /* Overview information follows */
            conn->active_method = nntpconnection_process_xover_data;

            if (more_data_ready) {
                conn->active_method(conn);
            }

            return;

        default:
            /* Unexpected response code. Abort the connection */
            nntpconnection_utils_strip_newline(buf, NULL);
            nntpconnection_disconnect_from_server(conn, DISCONNECT_INVALID_MSG, buf, __FILE__, __LINE__);
            return;
    }
}

static void
nntpconnection_process_xover_data(NntpConnection *conn)
{
#if 0
    char buf[1024];
    int length = 0;
    gboolean more_data_ready = FALSE;
    char **parts;
    struct tm tm;
    time_t post_date;

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->job_type == NNTP_JOB_TYPE_ARTICLE);
    g_return_if_fail(conn->part != NULL);
    g_return_if_fail(conn->article_fd >= 0);

    memset(buf, 0, sizeof(buf));

    /* Keep on reading data until we've found the '\r\n.\r\n' or '\n.\n' sequence */

    if (!nntpconnection_read_msg(conn, TRUE , sizeof(buf) - 1, buf, &length, &more_data_ready)) {
        return;
    }

    g_return_if_fail(length > 0);

    /* Are we at the end? */
    if (buf[0] == '.' && buf[1] == '\0') {
        conn->job_type = NNTP_JOB_TYPE_NONE;
        memset(&conn->newsgroup, 0, sizeof(conn->newsgroup));
        conn->active_method = NULL;

        return;
    }

    parts = g_strsplit(buf, "\t", 0);

    /* Sanity check */
    if (!parts    || !parts[0] || !parts[1] || !parts[2] || !parts[3] ||
        !parts[4] || !parts[5] || !parts[6] || !parts[7]) {

        goto out;
    }

    memset(&tm, 0, sizeof(tm));
    if (strptime(parts[3], "%d %b %Y %H:%M:%S %Z", &tm)) {
        post_date = mktime(&tm);
    } else {
        memset(&tm, 0, sizeof(tm));
        if (strptime(parts[3], "%A, %d %b %Y %H:%M:%S %Z", &tm)) {
            post_date = mktime(&tm);
        } else {
            // Date could not be parsed
            post_date = 0;

            goto out;
        }
    }

    //imported_funcs.parse_header(atoi(parts[0]), parts[1], parts[2], post_date, parts[4], atoi(parts[6]), atoi(parts[7]), conn->xover_start_range, conn->xover_end_range, data);

out:
    g_strfreev(parts);

    /* Process more data if it's available */
    if (more_data_ready) {
        conn->active_method(conn);
    }

    return;
#endif
}

void
nntpconnection_process_socket_activity(NntpConnection *conn)
{
    g_return_if_fail(conn != NULL);

    if (!conn->active_method) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "NNTP connection to server '%s' (connection_id %i, poll_fd %i) is in an undefined state", conn->server_info.servername, conn->connection_id, (int) conn->poll_fd.fd);
        nntpconnection_disconnect_from_server(conn, DISCONNECT_UNEXPECTED, _("NNTP Connection is in an undefined state"), __FILE__, __LINE__);
        return;
    }

    conn->active_method(conn);
    conn->last_activity_stamp = time(NULL);
}
