/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <wspiapi.h>
#include <io.h>
#undef  gai_strerror
#define gai_strerror gai_strerrorA
#define MSG_DONTWAIT 0
#else
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#ifdef HAVE_SSL
#ifdef HAVE_GNUTLS
#include <gnutls/gnutls.h>
#elif defined(HAVE_NSS_COMPAT_OSSL)
#include <nss_compat_ossl/nss_compat_ossl.h>
#else
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif
#endif

#include "nntpconnection_backend_native.h"
#include "nntpconnection_utils.h"
#include "nntpgrab_plugin.h"
#include "download_thread.h"

#if !GLIB_CHECK_VERSION(2,20,0) && !defined(WIN32)
/* The g_poll function was added in GLib 2.20.0.
 * Provide an implementation of this function for environments
 * which have an older version of GLib installed.
 * Source: http://git.gnome.org/browse/glib/tree/glib/gpoll.c
 */
#include <poll.h>

static gint
g_poll (GPollFD *fds, guint nfds, gint timeout)
{
    return poll ((struct pollfd *)fds, nfds, timeout);
}
#endif

#ifdef WIN32
/* The g_poll function only works on HANDLE's on Win32
 * so we replace the GLib implementation with one that
 * uses the select() function instead of poll().
 * The code below is taken from glib's gpoll.c where
 * it is a fallback path for platforms which don't
 * have a poll function() */
#undef g_poll
#define g_poll win32_g_poll
static gint
win32_g_poll (GPollFD *fds, guint nfds, gint timeout)
{
    struct timeval tv;
    FD_SET rset, wset, xset;
    GPollFD *f;
    int ready;
    int maxfd = 0;

    FD_ZERO (&rset);
    FD_ZERO (&wset);
    FD_ZERO (&xset);

    for (f = fds; f < &fds[nfds]; ++f)
        if (f->fd >= 0)
        {
            if (f->events & G_IO_IN)
                FD_SET (f->fd, &rset);
            if (f->events & G_IO_OUT)
                FD_SET (f->fd, &wset);
            if (f->events & G_IO_PRI)
                FD_SET (f->fd, &xset);
            if (f->fd > maxfd && (f->events & (G_IO_IN|G_IO_OUT|G_IO_PRI)))
                maxfd = f->fd;
        }

    tv.tv_sec = timeout / 1000;
    tv.tv_usec = (timeout % 1000) * 1000;

    ready = select (maxfd + 1, &rset, &wset, &xset,
                    timeout == -1 ? NULL : &tv);
    if (ready > 0)
        for (f = fds; f < &fds[nfds]; ++f)
        {
            f->revents = 0;
            if (f->fd >= 0)
            {
                if (FD_ISSET (f->fd, &rset))
                    f->revents |= G_IO_IN;
                if (FD_ISSET (f->fd, &wset))
                    f->revents |= G_IO_OUT;
                if (FD_ISSET (f->fd, &xset))
                    f->revents |= G_IO_PRI;
            }
        }

    return ready;
}
#endif

struct _NntpconnectionBackendNativePrivate {
    char recv_buffer[65500];    /* NOTE: private GObject fields may only be 64K in total */
    int recv_buffer_length;

#ifdef HAVE_SSL
#ifdef HAVE_GNUTLS
    gnutls_session_t ssl_session;
#else
    SSL *ssl;
#endif
#endif
};

struct _NntpconnectionBackendNativeClass {
    GObjectClass parent_class;
    /* define vtable methods and signals here */
};

static void nntpconnection_backend_native_iface_init(NntpconnectionBackendIface *iface);

static void nntpconnection_backend_native_init (NntpconnectionBackendNative *self)
{
    self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self, NNTPCONNECTION_TYPE_BACKEND_NATIVE, NntpconnectionBackendNativePrivate);
}

static void nntpconnection_backend_native_class_init (NntpconnectionBackendNativeClass *klass)
{
    g_type_class_add_private (klass, sizeof (NntpconnectionBackendNativePrivate));
}

G_DEFINE_TYPE_WITH_CODE (NntpconnectionBackendNative, nntpconnection_backend_native, NNTP_TYPE_CONNECTION,
                         G_IMPLEMENT_INTERFACE(NNTPCONNECTION_TYPE_BACKEND, nntpconnection_backend_native_iface_init))

#ifdef WIN32
static char *
WSAGetStrError(int err)
{
    LPVOID lpMsgBuf;
    char *ret;

    FormatMessageA(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        (DWORD) err,
        0,
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    ret = g_strdup(lpMsgBuf);
    nntpconnection_utils_strip_newline(ret, NULL);

    LocalFree(lpMsgBuf);

    return ret;
}
#endif

static gboolean
perform_recv(NntpConnection *conn, char **errmsg)
{
    NntpconnectionBackendNativePrivate *conn_priv = NNTPCONNECTION_BACKEND_NATIVE(conn)->priv;
    int retval;

    g_return_val_if_fail(conn != NULL, FALSE);

#ifdef HAVE_SSL
#ifdef HAVE_GNUTLS
    if (conn->server_info.use_ssl) {
        retval = gnutls_record_recv(conn_priv->ssl_session, conn_priv->recv_buffer + conn_priv->recv_buffer_length, sizeof(conn_priv->recv_buffer) - conn_priv->recv_buffer_length - 1);
        if (retval < 0) {
            switch (retval) {
                case GNUTLS_E_INTERRUPTED:
                case GNUTLS_E_AGAIN:
                    retval = -1;
#ifdef WIN32
                    WSASetLastError(WSAEWOULDBLOCK);
#else
                    errno = EAGAIN;
#endif
                    break;

                default:
                    g_print(__FILE__ ":%i gnutls_record_recv returned unknown value: %s (%i)\n", __LINE__, gnutls_strerror(retval), retval);
                    retval = 0;
                    break;
            }
        }
    } else
#else
    if (conn_priv->ssl) {
        int len = MIN(sizeof(conn_priv->recv_buffer) - conn_priv->recv_buffer_length - 1, 16384);     /* SSL packets are at most 16KB large */
        retval = SSL_read(conn_priv->ssl, conn_priv->recv_buffer + conn_priv->recv_buffer_length, len);
        if (retval < 0) {
            switch (SSL_get_error(conn_priv->ssl, retval)) {
                case SSL_ERROR_WANT_READ:
                case SSL_ERROR_WANT_WRITE:
#ifdef HAVE_NSS_COMPAT_OSSL
                case PR_PENDING_INTERRUPT_ERROR:
#endif
                    retval = -1;
#ifdef WIN32
                    WSASetLastError(WSAEWOULDBLOCK);
#else
                    errno = EAGAIN;
#endif
                    break;

                case SSL_ERROR_SSL:
                    retval = -1;
#ifdef WIN32
                    errno = ENOSYS;
#else
                    errno = EPROTO;
#endif
                    break;

                default:
                    g_print("Unknown SSL error detected on socket %i: %i\n", conn->poll_fd.fd, SSL_get_error(conn_priv->ssl, retval));
                    break;
            }
        }
    } else
#endif
#endif
    {
        retval = recv(conn->poll_fd.fd, conn_priv->recv_buffer + conn_priv->recv_buffer_length, sizeof(conn_priv->recv_buffer) - conn_priv->recv_buffer_length - 1, MSG_DONTWAIT);
    }

    switch (retval) {
        case -1:            /* Some error occured */
#ifdef WIN32
            if (WSAGetLastError() == WSAEWOULDBLOCK) {
#else
            if (errno == EAGAIN) {
#endif
                /* Try again right now */
                //g_print("EAGAIN detected on socket %i (SSL = %i)\n", conn->poll_fd.fd, conn->server_info.use_ssl);
                //return perform_recv(conn);
                return FALSE;
            }

#ifdef WIN32
            *errmsg = WSAGetStrError(WSAGetLastError());
#else
            *errmsg = g_strdup(strerror(errno));
#endif

            return FALSE;

        case 0:             /* Orderly shutdown has occured */
            conn->active_method = NULL;
            return FALSE;

        default:
            conn_priv->recv_buffer_length += retval;

            nntpconnection_backend_perform_shaping(conn->max_bandwidth, retval);

#ifdef HAVE_GNUTLS
            if (!conn->server_info.use_ssl)
#endif
                nntpconnection_backend_update_traffic_monitor(retval);

            break;
    }

//    nntpgrab_core_emit_debug_message(FALSE, "Received data from socket %i: %s", conn->connection_id, data);

    return TRUE;
}

static gboolean
nntpconnection_backend_native_ensure_data_is_ready(NntpConnection *conn, int timeout)
{
    g_return_val_if_fail(conn != NULL, FALSE);
    g_return_val_if_fail(timeout >= 0, FALSE);

    conn->poll_fd.events = G_IO_IN | G_IO_HUP | G_IO_ERR;
    conn->poll_fd.revents = 0;

    if (timeout > 0) {
        timeout /= 1000;        /* g_poll expects milliseconds, not microseconds */
    }

    while (TRUE) {
        int retval = g_poll(&conn->poll_fd, 1, timeout);

        switch (retval) {
            case -1:    /* Call was interrupted, see errno for details */
                if (errno == EINTR) {
                    continue;   /* Try again */
                }

#ifdef WIN32
                char *errmsg = WSAGetStrError(WSAGetLastError());
                ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_WARNING, __FILE__ ":%i Unknown error occured while polling: %s\n", __LINE__, errmsg);
                g_free(errmsg);
#else
                ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_WARNING, __FILE__ ":%i Unknown error occured while polling: %s\n", __LINE__, strerror(errno));
#endif

                return FALSE;

            case 0:     /* Timeout */
                return FALSE;

            default:
                return TRUE;
        }
    }

}

static gboolean
nntpconnection_backend_native_read_msg(NntpConnection *conn, gboolean read_line, int max_length, void *data, int *data_length, gboolean *more_data_ready, gboolean *try_again, char **errmsg)
{
    NntpconnectionBackendNativePrivate *conn_priv = NNTPCONNECTION_BACKEND_NATIVE(conn)->priv;
    int len;
    char *newline;

    g_return_val_if_fail(conn != NULL, FALSE);
    g_return_val_if_fail(max_length > 0, FALSE);
    g_return_val_if_fail(data != NULL, FALSE);
    g_return_val_if_fail(data_length != NULL, FALSE);
    g_return_val_if_fail(*data_length == 0, FALSE);
    /* more_data_ready can be NULL */
    g_return_val_if_fail(try_again != NULL, FALSE);
    g_return_val_if_fail(errmsg != NULL, FALSE);
    g_return_val_if_fail(*errmsg == NULL, FALSE);

    *try_again = FALSE;

    if (more_data_ready) {
        *more_data_ready = FALSE;
    }

    /* Avoid a buffer overflow */
    /* This situation can happen when a line is really long */
    if (conn_priv->recv_buffer_length >= sizeof(conn_priv->recv_buffer) - 1) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "nntpconnection_read_msg: Very long line received from server. Ignoring");
        memset(conn_priv->recv_buffer, 0, sizeof(conn_priv->recv_buffer));
        conn_priv->recv_buffer_length = 0;
    }

    /* Only read from the socket when there's no newline in the buffer */
    if (!(newline = memchr(conn_priv->recv_buffer, '\n', conn_priv->recv_buffer_length))) {
        if (!perform_recv(conn, errmsg)) {
            *try_again = TRUE;
            return FALSE;
        }

        /* Does the buffer contain a newline now? */
        if (!(newline = memchr(conn_priv->recv_buffer, '\n', conn_priv->recv_buffer_length))) {
            /* Try again later */
            *try_again = TRUE;
            return FALSE;
        }
    }

    g_return_val_if_fail(conn_priv->recv_buffer_length > 0, FALSE);

    /* Calculate the size of the buffer which needs to be sent back to the caller */
    if (!newline) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "nntpconnection_read_msg: Expected newline but didn't find any. contents = %s", conn_priv->recv_buffer);
        return FALSE;
    }

    len = newline - conn_priv->recv_buffer + 1;
    if (len > max_length) {
        /* We're going to have to trim some data */
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "nntpconnection_read_msg: Buffer isn't large enough, some data will be trimmed. contents = %s", conn_priv->recv_buffer);
        len = max_length;
    }

    memcpy(data, conn_priv->recv_buffer, len);

    if (len == conn_priv->recv_buffer_length) {
        /* Everything in the buffer was sent */
        conn_priv->recv_buffer_length = 0;
        memset(conn_priv->recv_buffer, 0, sizeof(conn_priv->recv_buffer));
    } else {
        /* Only part of the buffer was sent */
        memmove(conn_priv->recv_buffer, conn_priv->recv_buffer + len, sizeof(conn_priv->recv_buffer) - len);
        conn_priv->recv_buffer_length -= len;
        conn_priv->recv_buffer[conn_priv->recv_buffer_length] = '\0';

        g_return_val_if_fail(conn_priv->recv_buffer_length >= 0, FALSE);

        if (more_data_ready) {
            *more_data_ready = (memchr(conn_priv->recv_buffer, '\n', conn_priv->recv_buffer_length) != NULL);
        }
    }

    if (read_line) {
        nntpconnection_utils_strip_newline(data, &len);
    }

    if (data_length) {
        *data_length = len;
    }

    return TRUE;
}

static gboolean
nntpconnection_backend_native_send_msg(NntpConnection *conn, void *data, int length, char **errmsg)
{
    NntpconnectionBackendNativePrivate *conn_priv;
    int retval;
    int bytes_sent = 0;

    g_return_val_if_fail(conn != NULL, FALSE);
    g_return_val_if_fail(data != NULL, FALSE);
    g_return_val_if_fail(length > 0, FALSE);
    g_return_val_if_fail(errmsg != NULL, FALSE);
    g_return_val_if_fail(*errmsg == NULL, FALSE);

    conn_priv = NNTPCONNECTION_BACKEND_NATIVE(conn)->priv;
    g_return_val_if_fail(conn_priv != NULL, FALSE);

    if (length <= 0) {
        length = strlen(data);
    }

    do {
#ifdef HAVE_SSL
#ifdef HAVE_GNUTLS
        if (conn->server_info.use_ssl) {
            retval = gnutls_record_send(conn_priv->ssl_session, data + bytes_sent, length - bytes_sent);
        } else
#else
        if (conn_priv->ssl) {
            retval = SSL_write(conn_priv->ssl, data + bytes_sent, length - bytes_sent);
            if (retval < 0) {
                switch (SSL_get_error(conn_priv->ssl, retval)) {
                    case SSL_ERROR_WANT_READ:
                    case SSL_ERROR_WANT_WRITE:
                        retval = -1;
#ifdef WIN32
                        WSASetLastError(WSAEWOULDBLOCK);
#else
                        errno = EAGAIN;
#endif
                        break;

                    case SSL_ERROR_SSL:
                        retval = -1;
#ifdef WIN32
                        errno = ENOSYS;
#else
                        errno = EPROTO;
#endif
                        break;

                    default:
                        break;
                }
            }
        } else
#endif
#endif
        {
            retval = send(conn->poll_fd.fd, data + bytes_sent, length - bytes_sent, 0);
        }

        switch (retval) {
            case -1:            /* Some error occured */
            case 0:             /* Orderly shutdown has occured */
#ifdef WIN32
                *errmsg = WSAGetStrError(WSAGetLastError());
#else
                *errmsg = g_strdup(strerror(errno));
#endif
                return FALSE;

            default:
                bytes_sent += retval;

                if (bytes_sent == length) {
                    return TRUE;
                }

                break;
        }
    } while (TRUE);

    g_return_val_if_reached(TRUE);
}

/*****************************/
/* dns cache                 */
/*****************************/
struct _dns_cache_entry {
    char hostname[128];
    int port;
    struct addrinfo *res;
    int n;
};

static int
resolve_domain_name(const char *hostname, int port, struct addrinfo **res)
{
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
    static GList *cache = NULL;
    GList *list;
    struct addrinfo hints;
    int n = 0;
    char str_port[16];
    struct _dns_cache_entry *entry;

    g_return_val_if_fail(hostname != NULL, EAI_FAIL);
    g_return_val_if_fail(port > 0, EAI_FAIL);
    g_return_val_if_fail(port <= 65535, EAI_FAIL);
    g_return_val_if_fail(res != NULL, EAI_FAIL);

    *res = NULL;

    /* Search the DNS cache for a match first */
    g_static_mutex_lock(&mutex);
    list = cache;
    while (list) {
        entry = list->data;

        if (!strcmp(entry->hostname, hostname) && entry->port == port) {
            *res = entry->res;
            n = entry->n;

            g_static_mutex_unlock(&mutex);

            return n;
        }

        list = g_list_next(list);
    }

    /* DNS entry wasn't found in the cache yet. Perform a DNS resolve */
    entry = g_slice_new0(struct _dns_cache_entry);
    strncpy(entry->hostname, hostname, sizeof(entry->hostname) -1);
    entry->port = port;
    cache = g_list_append(cache, entry);

    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    memset(&str_port, 0, sizeof(str_port));
    snprintf(str_port, sizeof(str_port) - 1, "%i", port);
    entry->n = getaddrinfo(hostname, str_port, &hints, &entry->res);

    *res = entry->res;
    n = entry->n;

    g_static_mutex_unlock(&mutex);

    return n;
}

/*****************************/
/* generic socket operations */
/*****************************/
#ifdef HAVE_SSL

#ifdef HAVE_GNUTLS
static ssize_t pull_func (gnutls_transport_ptr_t fd, void *buf, size_t len)
{
    ssize_t retval = recv(GPOINTER_TO_INT(fd), buf, len, MSG_DONTWAIT);
    nntpconnection_backend_update_traffic_monitor(retval);
    return retval;
}

#ifndef GNUTLS_VERSION_MAJOR
#define GNUTLS_VERSION_MAJOR LIBGNUTLS_VERSION_MAJOR
#endif

#ifndef GNUTLS_VERSION_MINOR
#define GNUTLS_VERSION_MINOR LIBGNUTLS_VERSION_MINOR
#endif

#ifndef GNUTLS_VERSION_MICRO
#define GNUTLS_VERSION_MICRO LIBGNUTLS_VERSION_MICRO
#endif

#define GNUTLS_CHECK_VERSION(major,minor,micro)    \
    (GNUTLS_VERSION_MAJOR > (major) || \
     (GNUTLS_VERSION_MAJOR == (major) && GNUTLS_VERSION_MINOR > (minor)) || \
     (GNUTLS_VERSION_MAJOR == (major) && GNUTLS_VERSION_MINOR == (minor) && \
      GNUTLS_VERSION_MICRO >= (micro)))

static gboolean
prepare_gnutls_connection(NntpConnection *conn, char **errmsg)
{
    NntpconnectionBackendNativePrivate *conn_priv = NNTPCONNECTION_BACKEND_NATIVE(conn)->priv;
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
    static gboolean been_here = FALSE;
    gnutls_certificate_credentials_t cred;
    int ret;

    g_static_mutex_lock(&mutex);
    if (!been_here) {
        gnutls_global_init();
        been_here = TRUE;
    }
    g_static_mutex_unlock(&mutex);

    /* Initialize TLS session */
    gnutls_init (&conn_priv->ssl_session, GNUTLS_CLIENT);

    /* Use default priorities */
#if GNUTLS_CHECK_VERSION(2,1,7)
    gnutls_priority_set_direct (conn_priv->ssl_session, "PERFORMANCE:+ANON-DH:!ARCFOUR-128", NULL);
#endif

    /* put the anonymous credentials to the current session */
    gnutls_certificate_allocate_credentials(&cred);
    gnutls_credentials_set(conn_priv->ssl_session, GNUTLS_CRD_CERTIFICATE, cred);

    gnutls_transport_set_ptr (conn_priv->ssl_session, (gnutls_transport_ptr_t) GINT_TO_POINTER(conn->poll_fd.fd));

    /* Perform the TLS handshake */
    do {
        ret = gnutls_handshake (conn_priv->ssl_session);
    } while (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);

    if (ret < 0) {
        if (errmsg) {
            *errmsg = g_strdup_printf("Handshake failed: %s", gnutls_strerror(ret));
        }
        return FALSE;
    }

    /* Override the read function so that we can monitor the traffic on socket level */
    gnutls_transport_set_pull_function(conn_priv->ssl_session, pull_func);
    //gnutls_transport_set_ptr2 (conn->ssl_session, (gnutls_transport_ptr_t) pull_func, (gnutls_transport_ptr_t) push_func);

    return TRUE;
}
#else
static SSL *
prepare_ssl_connection(int conn_id, char **errmsg)
{
    SSL_CTX *ctx;
    SSL *ssl;
    SSL_METHOD *meth;
    int err;
#if 0
    char *str;
    X509 *server_cert;
#endif
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

    g_static_mutex_lock(&mutex);

    SSLeay_add_ssl_algorithms();
    meth = SSLv23_client_method();
    SSL_load_error_strings();
    g_static_mutex_unlock(&mutex);

    ctx = SSL_CTX_new (meth);
    if (!ctx) {
        if (errmsg) {
            *errmsg = g_strdup("SSL_CTX_new FAILED");
        }

        return NULL;
    }

    /* Don't verify the certificate as several usenet providers (like Eweka) use self-signed certificates */
    SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);

    ssl = SSL_new (ctx);
    SSL_CTX_free(ctx);

    if (!ssl) {
        if (errmsg) {
            *errmsg = g_strdup("SSL_new FAILED");
        }

        return NULL;
    }

    SSL_set_fd (ssl, conn_id);
    err = SSL_connect (ssl);
    if (err <= 0) {
        if (errmsg) {
            *errmsg = g_strdup_printf("%s", ERR_error_string(ERR_get_error(), NULL));
        }

        return NULL;
    }

#if 0
    printf ("SSL connection using %s\n", SSL_get_cipher (ssl));
    server_cert = SSL_get_peer_certificate (ssl);
    if (server_cert) {
        printf ("Server certificate:\n");

        str = X509_NAME_oneline (X509_get_subject_name (server_cert),0,0);
        if (str) {
            printf ("\t subject: %s\n", str);
            OPENSSL_free (str);
        }

        str = X509_NAME_oneline (X509_get_issuer_name (server_cert),0,0);
        if (str) {
            printf ("\t issuer: %s\n", str);
            OPENSSL_free (str);
        }

        /* We could do all sorts of certificate verification stuff here before deallocating the certificate. */

        X509_free (server_cert);
    }
#endif

    return ssl;
}
#endif
#endif

static NNTPConnectionErrCode
nntpconnection_backend_native_connect(NntpConnection *conn, const char *hostname, int port, char **errmsg)
{
    NntpconnectionBackendNativePrivate *conn_priv = NNTPCONNECTION_BACKEND_NATIVE(conn)->priv;
    int last_errno;
    struct addrinfo *res = NULL;
    int n;
#ifdef WIN32
    int tv;
    u_long ioctlArg;
#else
    struct timeval tv;
    int mode;
#endif
    NNTPDisconnectType disconnect_type;

    g_return_val_if_fail(conn != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(conn->poll_fd.fd == -1, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(hostname != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(port > 0, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(conn_priv != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(errmsg != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(*errmsg == NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);

    n = resolve_domain_name(hostname, port, &res);
    if (n != 0) {
        *errmsg = g_strdup(gai_strerror(n));
        return NNTP_CONNECTION_ERROR_HOST_NOT_FOUND;
    }

    last_errno = 0;

    while (res) {
        int ret;

        conn->poll_fd.fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);

#ifdef WIN32
        if ((conn->poll_fd.fd == INVALID_SOCKET)) {
#else
        if ((conn->poll_fd.fd == -1)) {
#endif
            // The socket couldn't be created
            // Save the errno and try the next item in the list
#ifdef WIN32
            last_errno = WSAGetLastError();
#else
            last_errno = errno;
#endif
            res = res->ai_next;
            conn->poll_fd.fd = -1;

            if (*errmsg) {
                g_free(*errmsg);
            }

#ifdef WIN32
            *errmsg = WSAGetStrError(last_errno);
#else
            *errmsg = g_strdup(strerror(last_errno));
#endif

            continue;
        }

        // Set the connection timeout on the socket
#ifdef WIN32
        tv = 5000;   /* F*cking Winsock uses an integer for the timeout and it's value is in milliseconds */

        setsockopt(conn->poll_fd.fd, SOL_SOCKET, SO_RCVTIMEO, (char*) &tv, sizeof(tv));
        setsockopt(conn->poll_fd.fd, SOL_SOCKET, SO_SNDTIMEO, (char*) &tv, sizeof(tv));
#else
        tv.tv_sec = 5;
        tv.tv_usec = 0;

        setsockopt(conn->poll_fd.fd, SOL_SOCKET, SO_RCVTIMEO, (const void*) &tv, sizeof(tv));
        setsockopt(conn->poll_fd.fd, SOL_SOCKET, SO_SNDTIMEO, (const void*) &tv, sizeof(tv));
#endif

        // Mark the socket as non-blocking
#ifdef WIN32
        ioctlArg = 1;
        ioctlsocket(conn->poll_fd.fd, FIONBIO, &ioctlArg);
#else
        mode = fcntl(conn->poll_fd.fd, F_GETFL, 0);
        fcntl(conn->poll_fd.fd, F_SETFL, mode | O_NONBLOCK);
#endif

        // Try to connect to the server
        ret = connect(conn->poll_fd.fd, res->ai_addr, (int) res->ai_addrlen);
        if (ret == 0) {
            // Connection succesfull
            last_errno = 0;
            break;
#ifndef WIN32
        } else if (errno == EINPROGRESS) {
            // Wait for at most 5 seconds
            socklen_t len;

            time_t now = time(NULL);
            do {
                // According to http://cr.yp.to/docs/connect.html there are various
                // ways to detect if a socket is connected. The getpeername() method
                // should be the most portable one
                if (res->ai_family == AF_INET6) {
                    struct sockaddr_in6 name;
                    len = sizeof(name);
                    if (getpeername(conn->poll_fd.fd, (struct sockaddr*) &name, &len) == 0) {
                        errno = 0;
                        break;
                    }
                } else {
                    struct sockaddr_in name;
                    len = sizeof(name);
                    if (getpeername(conn->poll_fd.fd, (struct sockaddr*) &name, &len) == 0) {
                        errno = 0;
                        break;
                    }
                }

                g_usleep(G_USEC_PER_SEC / 10);

                if (download_thread_get_state() != SCHEDULAR_STATE_RUNNING) {
                    break;
                }
            } while (now + 5 > time(NULL));

            if (errno == 0) {
                last_errno = 0;
                break;
            }
#else
        } else if (WSAGetLastError() == WSAEWOULDBLOCK) {
            // And Windows again has different behaviour....
            struct timeval tv;
            fd_set send_fds;
            int len;

            FD_ZERO(&send_fds);
            FD_SET(conn->poll_fd.fd, &send_fds);

            tv.tv_sec = 5;
            tv.tv_usec = 0;

            len = select(conn->poll_fd.fd + 1, NULL, &send_fds, NULL, &tv);
            if (len <= 0) {
                // Read timeout or some other error
                if (len == 0) {
                    errno = WSAETIMEDOUT;
                }
            } else {
                // Connection is made
                last_errno = 0;
                break;
            }
#endif
        }

        // Connection could not be made, save the errno
#ifndef WIN32
        if (errno == EINPROGRESS) {
            // The connection could not be established within the connect time limit.
            // This provides a more clear error message
            last_errno = ETIMEDOUT;
        } else
#endif
        {
            last_errno = errno;
        }

        if (*errmsg) {
            g_free(*errmsg);
        }

#ifdef WIN32
        *errmsg = WSAGetStrError(last_errno);
#else
        *errmsg = g_strdup(strerror(last_errno));
#endif

#if WIN32
        if (last_errno == WSAETIMEDOUT) {
#else
        if (last_errno == ETIMEDOUT) {
#endif
            disconnect_type = DISCONNECT_CONNECT_TIMEOUT;
        } else {
            disconnect_type = DISCONNECT_CONNECTION_REFUSED;
        }

#ifdef WIN32
        nntpconnection_disconnect_from_server(conn, disconnect_type, WSAGetStrError(last_errno), __FILE__, __LINE__);
#else
        nntpconnection_disconnect_from_server(conn, disconnect_type, strerror(last_errno), __FILE__, __LINE__);
#endif

        res = res->ai_next;
    }

    if (conn->poll_fd.fd == -1) {
        switch (last_errno) {
#if WIN32
            case WSAETIMEDOUT:
#else
            case ETIMEDOUT:
#endif
                return NNTP_CONNECTION_ERROR_CONNECTION_TIMEOUT;

            default:
                return NNTP_CONNECTION_ERROR_CONNECTION_REFUSED;
        };
    }

    /* Mark the socket as blocking */
#ifdef WIN32
    ioctlArg = 0;
    ioctlsocket(conn->poll_fd.fd, FIONBIO, &ioctlArg);
#else
    mode = fcntl(conn->poll_fd.fd, F_GETFL, 0);
    fcntl(conn->poll_fd.fd, F_SETFL, mode  ^ O_NONBLOCK);
#endif

    memset(conn_priv->recv_buffer, 0, sizeof(conn_priv->recv_buffer));
    conn_priv->recv_buffer_length = 0;

    return NNTP_CONNECTION_ERROR_NONE;
}

static void
nntpconnection_backend_native_disconnect(NntpConnection *conn)
{

#ifdef WIN32
    /* DisconnectEx(conn->poll_fd.fd, NULL, 0, 0);   *  MinGW doesn't have DisconnectEx exported... */
    shutdown(conn->poll_fd.fd, SD_BOTH);            /* The shutdown() function is broken on Win32, see
                                                     * http://msdn.microsoft.com/en-us/library/ms738547%28VS.85%29.aspx
                                                     * for details*/
    closesocket(conn->poll_fd.fd);
#else
    shutdown(conn->poll_fd.fd, SHUT_RDWR);
    close(conn->poll_fd.fd);
#endif

    conn->poll_fd.fd = -1;
}

static gboolean
nntpconnection_backend_native_prepare_ssl (NntpConnection *conn, char **errmsg)
{
    NntpconnectionBackendNativePrivate *conn_priv = NNTPCONNECTION_BACKEND_NATIVE(conn)->priv;

    g_return_val_if_fail(conn != NULL, FALSE);
    g_return_val_if_fail(conn->server_info.use_ssl == TRUE, FALSE);
    g_return_val_if_fail(conn_priv != NULL, FALSE);

#ifdef HAVE_SSL
#ifdef HAVE_GNUTLS
    if (!prepare_gnutls_connection(conn, errmsg)) {
#else
    if ((conn_priv->ssl = prepare_ssl_connection(conn->poll_fd.fd, errmsg)) == NULL) {
#endif
        return FALSE;
    } else {
        return TRUE;
    }
#else
    g_assert_not_reached();
#endif
}

static gboolean
nntpconnection_backend_native_get_has_ssl_support(NntpConnection *conn)
{
#ifdef HAVE_SSL
    return TRUE;
#else
    return FALSE;
#endif
}

static void
nntpconnection_backend_native_iface_init(NntpconnectionBackendIface *iface)
{
    iface->connect = nntpconnection_backend_native_connect;
    iface->disconnect = nntpconnection_backend_native_disconnect;
    iface->prepare_ssl = nntpconnection_backend_native_prepare_ssl;
    iface->ensure_data_is_ready = nntpconnection_backend_native_ensure_data_is_ready;
    iface->read_msg = nntpconnection_backend_native_read_msg;
    iface->send_msg = nntpconnection_backend_native_send_msg;
    iface->get_has_ssl_support = nntpconnection_backend_native_get_has_ssl_support;
}

NntpConnection *
nntpconnection_backend_native_new (void)
{
    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Created NNTP connection using backend 'native'");
    return NNTP_CONNECTION(g_object_new(NNTPCONNECTION_TYPE_BACKEND_NATIVE, NULL));
}
