/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <nm-client.h>
#include "nntpgrab_plugin.h"
#include "config.h"

static void nm_state_changed_cb(GObject *gobject, GParamSpec *pspec, gpointer user_data);

struct _nm_plugin_data {
    NMClient *client;
    gboolean schedular_stopped_by_this_plugin;
};

void
nntpgrab_plugin_initialize(NGPlugin *plugin_data)
{
    ng_plugin_set_name(plugin_data, "NetworkManager");
    ng_plugin_set_version(plugin_data, PACKAGE_VERSION);
    ng_plugin_set_author(plugin_data, "Erik van Pienbroek");
    ng_plugin_set_url(plugin_data, "http://www.nntpgrab.nl");
    ng_plugin_set_description(plugin_data, "Plugin which offers integration with NetworkManager. Whenever a network connection is lost, the schedular will automatically get paused. When the network connection is resumed, the schedular will automatically be resumed");
}

ngboolean
nntpgrab_plugin_load(NGPlugin *plugin_data, char **errmsg)
{
    struct _nm_plugin_data *priv = g_slice_new0(struct _nm_plugin_data);
    priv->client = (gpointer) nm_client_new();
    priv->schedular_stopped_by_this_plugin = FALSE;
    plugin_data->priv = (gpointer) priv;

    g_signal_connect(G_OBJECT(priv->client), "notify::state", G_CALLBACK(nm_state_changed_cb), plugin_data);

    return TRUE;
}

ngboolean
nntpgrab_plugin_can_unload(NGPlugin *plugin_data, char **reason)
{
    return TRUE;
}

void
nntpgrab_plugin_unload(NGPlugin *plugin_data)
{
    if (plugin_data->priv) {
        g_object_unref(((struct _nm_plugin_data*) plugin_data->priv)->client);
        g_slice_free(struct _nm_plugin_data, plugin_data->priv);

        plugin_data->priv = NULL;
    }
}

void
nntpgrab_plugin_destroy(NGPlugin *plugin_data)
{
}

int
nntpgrab_plugin_get_version(void)
{
    return NNTPGRAB_PLUGIN_API_VERSION;
}

gboolean
start_schedular_cb(gpointer data)
{
    NGPlugin *plugin_data = (NGPlugin*) data;

    switch (nm_client_get_state(NM_CLIENT(((struct _nm_plugin_data*)plugin_data->priv)->client))) {
        case NM_STATE_CONNECTED:
        case NM_STATE_UNKNOWN:
            if (((struct _nm_plugin_data*)plugin_data->priv)->schedular_stopped_by_this_plugin &&
                plugin_data->core_funcs.schedular_get_state() == SCHEDULAR_STATE_STOPPED) {

                if (plugin_data->core_funcs.schedular_start() == FALSE) {
                    /* The schedular is probably still being paused, try again later */
                    return TRUE;
                }

                ((struct _nm_plugin_data*)plugin_data->priv)->schedular_stopped_by_this_plugin = FALSE;
            }

            break;

        case NM_STATE_ASLEEP:
        case NM_STATE_CONNECTING:
        case NM_STATE_DISCONNECTED:
            break;
    }

    return FALSE;
}

static void
nm_state_changed_cb(GObject *gobject, GParamSpec *pspec, gpointer user_data)
{
    NGPlugin *plugin_data = (NGPlugin*) user_data;

    g_print("NetworkManager state changed! new state = %i\n", nm_client_get_state(NM_CLIENT(gobject)));
    switch (nm_client_get_state(NM_CLIENT(gobject))) {
        case NM_STATE_ASLEEP:
        case NM_STATE_DISCONNECTED:
            /* Only stop the schedular when it is still running */
            if (plugin_data->core_funcs.schedular_get_state() == SCHEDULAR_STATE_RUNNING) {
                ((struct _nm_plugin_data*)plugin_data->priv)->schedular_stopped_by_this_plugin = TRUE;
                plugin_data->core_funcs.schedular_stop(NULL, FALSE);        /* We don't use the reason field here as we don't want to show the user a popup for this situation */
            }
            break;
        case NM_STATE_CONNECTED:
        case NM_STATE_UNKNOWN:
            /* Frequently it takes a few additional seconds to get the network fully up and running so we use a timer here */
#if GLIB_CHECK_VERSION(2,14,0)
            g_timeout_add_seconds(5, start_schedular_cb, plugin_data);
#else
            g_timeout_add(5000, start_schedular_cb, plugin_data);
#endif
            break;

        default:
        case NM_STATE_CONNECTING:
            break;
    }
}
